# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Tuple, Optional, Union
from dataclasses import dataclass
import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange
from transformers.activations import ACT2FN
from transformers import AutoConfig

try:
    from flash_attn.bert_padding import pad_input, unpad_input
    from flash_attn.flash_attn_interface import flash_attn_varlen_qkvpacked_func

    has_flash_attn = True
except Exception:
    print("FlashAttention2 is not installed.")
    has_flash_attn = False


@dataclass
class InternVL_Encoder_Args:
    num_channels: int
    patch_size: int
    image_size: int
    qkv_bias: bool
    hidden_size: int
    num_attention_heads: int
    intermediate_size: int
    qk_normalization: bool
    num_hidden_layers: int
    hidden_act: str
    norm_type: str
    layer_norm_eps: float = 1e-6
    hf_config: AutoConfig = None


class FlashAttention(nn.Module):
    """Implement the scaled dot product attention with softmax.
    Arguments
    ---------
        softmax_scale: The temperature to use for the softmax attention.
                      (default: 1/sqrt(d_keys) where d_keys is computed at
                      runtime)
        attention_dropout: The dropout rate to apply to the attention
                           (default: 0.0)
    """

    def __init__(self, softmax_scale=None):
        super().__init__()
        self.softmax_scale = softmax_scale

    def forward(
        self,
        qkv,
        key_padding_mask=None,
        causal=False,
        cu_seqlens=None,
        max_s=None,
        need_weights=False,
    ):
        """Implements the multihead softmax attention.
        Arguments
        ---------
            qkv: The tensor containing the query, key, and value. (B, S, 3, H, D) if key_padding_mask is None
                if unpadded: (nnz, 3, h, d)
            key_padding_mask: a bool tensor of shape (B, S)
        """
        assert not need_weights
        assert qkv.dtype in [torch.float16, torch.bfloat16]
        assert qkv.is_cuda

        if cu_seqlens is None:
            batch_size = qkv.shape[0]
            seqlen = qkv.shape[1]
            if key_padding_mask is None:
                qkv = rearrange(qkv, "b s ... -> (b s) ...")
                max_s = seqlen
                cu_seqlens = torch.arange(
                    0,
                    (batch_size + 1) * seqlen,
                    step=seqlen,
                    dtype=torch.int32,
                    device=qkv.device,
                )
                output = flash_attn_varlen_qkvpacked_func(
                    qkv,
                    cu_seqlens,
                    max_s,
                    0.0,
                    softmax_scale=self.softmax_scale,
                    causal=causal,
                )
                output = rearrange(output, "(b s) ... -> b s ...", b=batch_size)
            else:
                nheads = qkv.shape[-2]
                x = rearrange(qkv, "b s three h d -> b s (three h d)")
                x_unpad, indices, cu_seqlens, max_s = unpad_input(x, key_padding_mask)
                x_unpad = rearrange(
                    x_unpad, "nnz (three h d) -> nnz three h d", three=3, h=nheads
                )
                output_unpad = flash_attn_varlen_qkvpacked_func(
                    x_unpad,
                    cu_seqlens,
                    max_s,
                    0.0,
                    softmax_scale=self.softmax_scale,
                    causal=causal,
                )
                output = rearrange(
                    pad_input(
                        rearrange(output_unpad, "nnz h d -> nnz (h d)"),
                        indices,
                        batch_size,
                        seqlen,
                    ),
                    "b s (h d) -> b s h d",
                    h=nheads,
                )
        else:
            assert max_s is not None
            output = flash_attn_varlen_qkvpacked_func(
                qkv,
                cu_seqlens,
                max_s,
                0.0,
                softmax_scale=self.softmax_scale,
                causal=causal,
            )

        return output, None


class InternRMSNorm(nn.Module):
    def __init__(self, hidden_size, eps=1e-6):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(hidden_size))
        self.variance_epsilon = eps

    def forward(self, hidden_states):
        input_dtype = hidden_states.dtype
        hidden_states = hidden_states.to(torch.float32)
        variance = hidden_states.pow(2).mean(-1, keepdim=True)
        hidden_states = hidden_states * torch.rsqrt(variance + self.variance_epsilon)
        return self.weight * hidden_states.to(input_dtype)


NORM2FN = {
    "rms_norm": InternRMSNorm,
    "layer_norm": nn.LayerNorm,
}


class InternVisionEmbeddings(nn.Module):
    def __init__(self, config: InternVL_Encoder_Args):
        super().__init__()
        self.config = config
        self.embed_dim = config.hidden_size
        self.image_size = config.image_size
        self.patch_size = config.patch_size

        self.class_embedding = nn.Parameter(
            torch.randn(1, 1, self.embed_dim),
        )

        self.patch_embedding = nn.Conv2d(
            in_channels=3,
            out_channels=self.embed_dim,
            kernel_size=self.patch_size,
            stride=self.patch_size,
        )

        self.num_patches = (self.image_size // self.patch_size) ** 2
        self.num_positions = self.num_patches + 1

        self.position_embedding = nn.Parameter(
            torch.randn(1, self.num_positions, self.embed_dim)
        )

    def _get_pos_embed(self, pos_embed, H, W):
        target_dtype = pos_embed.dtype
        pos_embed = (
            pos_embed.float()
            .reshape(
                1,
                self.image_size // self.patch_size,
                self.image_size // self.patch_size,
                -1,
            )
            .permute(0, 3, 1, 2)
        )
        pos_embed = (
            F.interpolate(pos_embed, size=(H, W), mode="bicubic", align_corners=False)
            .reshape(1, -1, H * W)
            .permute(0, 2, 1)
            .to(target_dtype)
        )
        return pos_embed

    def forward(self, pixel_values: torch.FloatTensor) -> torch.Tensor:
        target_dtype = self.patch_embedding.weight.dtype
        patch_embeds = self.patch_embedding(
            pixel_values
        )  # shape = [*, channel, width, height]
        batch_size, _, height, width = patch_embeds.shape
        patch_embeds = patch_embeds.flatten(2).transpose(1, 2)
        class_embeds = self.class_embedding.expand(batch_size, 1, -1).to(target_dtype)
        embeddings = torch.cat([class_embeds, patch_embeds], dim=1)
        position_embedding = torch.cat(
            [
                self.position_embedding[:, :1, :],
                self._get_pos_embed(self.position_embedding[:, 1:, :], height, width),
            ],
            dim=1,
        )
        embeddings = embeddings + position_embedding.to(target_dtype)
        return embeddings


class InternAttention(nn.Module):
    """Multi-headed attention from 'Attention Is All You Need' paper"""

    def __init__(self, config: InternVL_Encoder_Args):
        super().__init__()
        self.config = config
        self.embed_dim = config.hidden_size
        self.num_heads = config.num_attention_heads
        self.head_dim = self.embed_dim // self.num_heads
        if self.head_dim * self.num_heads != self.embed_dim:
            raise ValueError(
                f"embed_dim must be divisible by num_heads (got `embed_dim`: {self.embed_dim} and `num_heads`:"
                f" {self.num_heads})."
            )

        self.scale = self.head_dim**-0.5
        self.qkv = nn.Linear(self.embed_dim, 3 * self.embed_dim, bias=config.qkv_bias)

        self.qk_normalization = config.qk_normalization

        if self.qk_normalization:
            self.q_norm = InternRMSNorm(self.embed_dim, eps=config.layer_norm_eps)
            self.k_norm = InternRMSNorm(self.embed_dim, eps=config.layer_norm_eps)

        self.inner_attn = FlashAttention()
        self.proj = nn.Linear(self.embed_dim, self.embed_dim)

    def _flash_attn(self, x, key_padding_mask=None, need_weights=False):
        qkv = self.qkv(x)
        qkv = rearrange(
            qkv, "b s (three h d) -> b s three h d", three=3, h=self.num_heads
        )

        if self.qk_normalization:
            q, k, v = qkv.unbind(2)
            q = self.q_norm(q.flatten(-2, -1)).view(q.shape)
            k = self.k_norm(k.flatten(-2, -1)).view(k.shape)
            qkv = torch.stack([q, k, v], dim=2)

        context, _ = self.inner_attn(
            qkv,
            key_padding_mask=key_padding_mask,
            need_weights=need_weights,
            causal=False,
        )
        outs = self.proj(rearrange(context, "b s h d -> b s (h d)"))
        return outs

    def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
        x = self._flash_attn(hidden_states)
        return x


class InternMLP(nn.Module):
    def __init__(self, config: InternVL_Encoder_Args):
        super().__init__()
        self.config = config
        self.act = ACT2FN[config.hidden_act]
        self.fc1 = nn.Linear(config.hidden_size, config.intermediate_size)
        self.fc2 = nn.Linear(config.intermediate_size, config.hidden_size)

    def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
        hidden_states = self.fc1(hidden_states)
        hidden_states = self.act(hidden_states)
        hidden_states = self.fc2(hidden_states)
        return hidden_states


class InternVisionEncoderLayer(nn.Module):
    def __init__(self, config: InternVL_Encoder_Args):
        super().__init__()
        self.embed_dim = config.hidden_size
        self.intermediate_size = config.intermediate_size
        self.norm_type = config.norm_type

        self.attn = InternAttention(config)
        self.mlp = InternMLP(config)
        self.norm1 = NORM2FN[self.norm_type](self.embed_dim, eps=config.layer_norm_eps)
        self.norm2 = NORM2FN[self.norm_type](self.embed_dim, eps=config.layer_norm_eps)

        self.ls1 = nn.Parameter(torch.ones(self.embed_dim))
        self.ls2 = nn.Parameter(torch.ones(self.embed_dim))

    def forward(
        self,
        hidden_states: torch.Tensor,
    ):
        """
        Args:
            hidden_states (`Tuple[torch.FloatTensor, Optional[torch.FloatTensor]]`): input to the layer of shape `(batch, seq_len, embed_dim)`
        """
        hidden_states = (
            hidden_states
            + self.attn(self.norm1(hidden_states).to(hidden_states.dtype)) * self.ls1
        )

        hidden_states = (
            hidden_states
            + self.mlp(self.norm2(hidden_states).to(hidden_states.dtype)) * self.ls2
        )

        return hidden_states


class InternVisionEncoder(nn.Module):
    """
    Transformer encoder consisting of `config.num_hidden_layers` self attention layers. Each layer is a
    [`InternEncoderLayer`].

    Args:
        config (`InternConfig`):
            The corresponding vision configuration for the `InternEncoder`.
    """

    def __init__(self, config: InternVL_Encoder_Args):
        super().__init__()
        self.config = config
        self.layers = torch.nn.ModuleDict()
        for layer_id in range(config.num_hidden_layers):
            self.layers[str(layer_id)] = InternVisionEncoderLayer(config)

    def forward(
        self,
        inputs_embeds,
        select_layer: Optional[int] = None,
    ) -> Union[Tuple, torch.Tensor]:
        r"""
        Args:
            inputs_embeds (`torch.FloatTensor` of shape `(batch_size, sequence_length, hidden_size)`):
                Embedded representation of the inputs. Should be float, not int tokens.
            output_hidden_states (`bool`, *optional*):
                Whether or not to return the hidden states of all layers. See `hidden_states` under returned tensors
                for more detail.
        """
        output_hidden_states = select_layer != -1
        encoder_states = () if output_hidden_states else None
        hidden_states = inputs_embeds

        for idx, encoder_layer in self.layers.items():
            if output_hidden_states:
                encoder_states = encoder_states + (hidden_states,)
            if (
                hasattr(encoder_layer, "_gradient_checkpointing_enabled")
                and encoder_layer._gradient_checkpointing_enabled
            ):
                layer_outputs = torch.utils.checkpoint.checkpoint(
                    encoder_layer, hidden_states
                )
            else:
                layer_outputs = encoder_layer(
                    hidden_states,
                )
            hidden_states = layer_outputs
        if output_hidden_states:
            encoder_states = encoder_states + (hidden_states,)
        return hidden_states if select_layer == -1 else encoder_states[select_layer]


class InternVisionModel(nn.Module):
    def __init__(self, config) -> None:
        super().__init__()
        self.config = config

        self.embeddings = InternVisionEmbeddings(config)
        self.encoder = InternVisionEncoder(config)

    def forward(
        self,
        pixel_values: Optional[torch.FloatTensor] = None,
        pixel_embeds: Optional[torch.FloatTensor] = None,
        select_layer: Optional[int] = None,
    ) -> Union[torch.Tensor, Tuple[torch.Tensor]]:
        if pixel_values is None and pixel_embeds is None:
            raise ValueError("You have to specify pixel_values or pixel_embeds")

        if pixel_embeds is not None:
            hidden_states = pixel_embeds
        else:
            if len(pixel_values.shape) == 4:
                hidden_states = self.embeddings(pixel_values)
            else:
                raise ValueError(f"wrong pixel_values size: {pixel_values.shape}")
        return self.encoder(
            inputs_embeds=hidden_states,
            select_layer=select_layer,
        )

    def _get_nparams_and_flops_fn(self) -> Callable[[int], tuple[int, int]]:
        nparams = sum(p.numel() for p in self.parameters())
        nparams_embedding = sum(
            sum(p.numel() for p in m.parameters())
            for m in self.children()
            if isinstance(m, nn.Embedding)
        )

        # Reasoning behind the factor of 12 for the self-attention part of the formula:
        # 1. each self-attention has 2 matmul in the forward and 4 in the backward (6)
        # 2. the flash attention does 1 more matmul recomputation in the backward
        #    but recomputation should not be counted in calculating MFU           (+0)
        # 3. each matmul performs 1 multiplication and 1 addition                 (*2)
        # 4. we follow the convention and do not account for sparsity in causal attention
        layers, heads, head_dim = (
            len(self.blocks),
            self.config.n_heads,
            self.config.hidden_size // self.config.n_heads,
        )
        return lambda seq_len: (
            nparams,
            6 * (nparams - nparams_embedding)
            + 12 * layers * heads * head_dim * seq_len,
        )



from cosmos_rl.dispatcher.data.packer import worker_entry_parser
from cosmos_rl.launcher.worker_entry import main as launch_worker
from cosmos_rl.tools.dataset.gsm8k_grpo import (
    GSM8kDataset,
    GSM8kValDataset,
    custom_reward_fn,
    custom_logger_fn,
    GSM8kDataPacker,
)
from cosmos_rl.utils.logging import logger
from cosmos_rl.policy.config import Config as CosmosConfig
from torch.utils.data import Dataset


class MinimalGSM8kDataset(GSM8kDataset):
    def __len__(self):
        return 8


class MinimalGSM8kValDataset(GSM8kValDataset):
    def __len__(self):
        return 8


if __name__ == "__main__":
    parser = worker_entry_parser()

    # Users can add custom arguments here
    parser.add_argument(
        "--foo", type=str, default="bar", help="The custom optional argument name."
    )
    parser.add_argument(
        "x_arg",
        type=str,
        default=None,
        help="The custom positional argument name.",
    )

    try:
        args = parser.parse_args()
        assert args.x_arg == "cosmos_rl"
        assert args.foo == "cosmos"
    except SystemExit as e:
        logger.error("Error parsing arguments.")
        raise e

    def get_dataset(config: CosmosConfig) -> Dataset:
        return MinimalGSM8kDataset()

    def get_val_dataset(config: CosmosConfig) -> Dataset:
        return MinimalGSM8kValDataset()

    launch_worker(
        dataset=get_dataset,
        val_dataset=get_val_dataset,
        reward_fns=[custom_reward_fn],
        data_packer=GSM8kDataPacker(),
        val_data_packer=GSM8kDataPacker(),
        custom_logger_fns=[custom_logger_fn],
        args=args,  # Note: args must be passed if you want to use custom arguments
    )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def cdiv(a: int, b: int) -> int:
    return -(a // -b)


def next_power_of_2(n) -> int:
    if n < 1:
        return 1
    return 1 << (n - 1).bit_length()


def prev_power_of_2(n: int) -> int:
    if n <= 0:
        return 0
    return 1 << (n.bit_length() - 1)


def round_up(x: int, y: int) -> int:
    return ((x + y - 1) // y) * y


def round_down(x: int, y: int) -> int:
    return (x // y) * y



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import copy
import torch
import unittest
from contextlib import contextmanager
from PIL import Image
from qwen_vl_utils import process_vision_info
from transformers import AutoConfig, AutoProcessor

from cosmos_rl.utils.parallelism import ParallelDims
from cosmos_rl.policy.model.hf_models import HFModel
from cosmos_rl.policy.config import Config as CosmosConfig, ParallelismConfig


@contextmanager
def cosmos_default_dtype(dtype: torch.dtype):
    old = torch.get_default_dtype()
    torch.set_default_dtype(dtype)
    try:
        yield
    finally:
        torch.set_default_dtype(old)


def test_hf_model_generate(model, inputs):
    generation_kwargs = dict(
        max_new_tokens=1,
        output_logits=True,
        return_dict_in_generate=True,
    )
    logits = model.generate(**inputs, **generation_kwargs).logits[0]
    return logits


def test_hf_model_forward(model, inputs):
    with torch.no_grad():
        logits = model(**inputs).logits
        return logits[:, -1, :]


def test_cosmos_hf_model(model, inputs):
    with torch.no_grad():
        logits = model(**inputs)
        return logits[:, -1, :]


class TestHFModel(unittest.TestCase):
    def test_post_to_empty_hook(self):
        for model_id in [
            "Qwen/Qwen2.5-VL-7B-Instruct",
            "llava-hf/llava-1.5-7b-hf",
            "Qwen/Qwen3-VL-4B-Instruct",
            # "google/gemma-3-12b-it", # Need access to the repo
            # "mistralai/Mistral-7B-Instruct-v0.3", # Need access to the repo
            "microsoft/phi-4",
        ]:
            for dtype in [torch.bfloat16, torch.float32]:
                # To avoid out-of-memory issues, bypass float32 precision for models which have more than 10B parameters
                if (
                    model_id in ["google/gemma-3-12b-it", "microsoft/phi-4"]
                    and dtype == torch.float32
                ):
                    continue
                max_position_embeddings = 1024
                # Load config
                config = AutoConfig.from_pretrained(model_id, trust_remote_code=True)
                config.max_position_embeddings = max_position_embeddings
                config.torch_dtype = dtype
                # Load cosmos hf model
                cosmos_hf_model = None
                with torch.device("meta"):
                    with cosmos_default_dtype(dtype):
                        cosmos_hf_model = HFModel.from_pretrained(
                            config,
                            model_id,
                            max_position_embeddings=max_position_embeddings,
                        )
                cosmos_hf_model.to_empty(device="cuda")
                cosmos_hf_model.post_to_empty_hook(CosmosConfig())
                parallel_dims = ParallelDims.from_config(ParallelismConfig(tp_size=1))
                cosmos_hf_model.load_hf_weights(
                    model_id, parallel_dims, "cuda", revision=None
                )

                # Load hf model
                hf_model = cosmos_hf_model.model_class.from_pretrained(
                    model_id, trust_remote_code=True, config=config
                ).to("cuda", dtype=dtype)
                hf_named_buffers = {k: v for k, v in hf_model.named_buffers()}

                for name, cosmos_hf_buffer in cosmos_hf_model.model.named_buffers():
                    assert (
                        name in hf_named_buffers
                    ), f"Buffer {name} not found in hf model"
                    hf_buffer = hf_named_buffers[name]
                    assert (
                        cosmos_hf_buffer.shape == hf_buffer.shape
                    ), f"Shape mismatch: {cosmos_hf_buffer.shape} != {hf_buffer.shape} for {name}"
                    assert (
                        cosmos_hf_buffer.dtype == hf_buffer.dtype
                    ), f"Dtype mismatch: {cosmos_hf_buffer.dtype} != {hf_buffer.dtype} for {name}"
                    assert torch.equal(
                        cosmos_hf_buffer, hf_buffer
                    ), f"Buffer {name} is not equal to the one in hf model"

                print(f"{model_id} with {dtype=} post_to_empty_hook test passed.")
                del cosmos_hf_model
                del hf_model
                del hf_named_buffers
                torch.cuda.empty_cache()

    def test_forward(self):
        for model_id in [
            "Qwen/Qwen2.5-VL-7B-Instruct",
            "llava-hf/llava-1.5-7b-hf",
            "Qwen/Qwen3-VL-4B-Instruct",
            # "google/gemma-3-12b-it", # Need access to the repo
            # "mistralai/Mistral-7B-Instruct-v0.3", # Need access to the repo
            "microsoft/phi-4",
        ]:
            dtype = torch.bfloat16
            max_position_embeddings = 4096
            config = AutoConfig.from_pretrained(model_id, trust_remote_code=True)
            config.max_position_embeddings = max_position_embeddings
            config._attn_implementation = "eager"
            config.torch_dtype = dtype
            cosmos_hf_model = None
            with torch.device("meta"):
                with cosmos_default_dtype(dtype):
                    cosmos_hf_model = HFModel.from_pretrained(
                        config,
                        model_id,
                        max_position_embeddings=max_position_embeddings,
                    )
            cosmos_hf_model.to_empty(device="cuda")
            cosmos_hf_model.post_to_empty_hook(CosmosConfig())
            parallel_dims = ParallelDims.from_config(ParallelismConfig(tp_size=1))
            cosmos_hf_model.load_hf_weights(
                model_id, parallel_dims, "cuda", revision=None
            )

            hf_model = cosmos_hf_model.model_class.from_pretrained(
                model_id, trust_remote_code=True, config=config
            ).to("cuda", dtype=dtype)
            processor = AutoProcessor.from_pretrained(
                model_id, trust_remote_code=True, use_fast=True
            )
            try:
                if cosmos_hf_model.is_vlm:
                    current_dir = os.path.dirname(os.path.abspath(__file__))
                    image = Image.open(
                        os.path.join(current_dir, "data", "test_hf_model.jpg")
                    )
                    messages = [
                        [
                            {
                                "role": "user",
                                "content": [
                                    {
                                        "type": "image",
                                        "image": image,
                                    },
                                    {"type": "text", "text": "describe the image"},
                                ],
                            }
                        ]
                    ]
                    text = processor.apply_chat_template(messages, tokenize=False)
                    image_inputs, video_inputs = process_vision_info(messages)

                    inputs = processor(
                        text=text,
                        images=image_inputs,
                        videos=video_inputs,
                        padding=True,
                        return_tensors="pt",
                    ).to("cuda")
                else:
                    messages = [
                        {
                            "role": "system",
                            "content": "You are a pirate chatbot who always responds in pirate speak!",
                        },
                        {"role": "user", "content": "Who are you?"},
                    ]
                    text = processor.apply_chat_template(messages, tokenize=False)
                    inputs = processor(
                        text=text,
                        return_tensors="pt",
                    ).to("cuda")

                hf_generate_logits = test_hf_model_generate(
                    hf_model, copy.deepcopy(inputs)
                )
                hf_forward_logits = test_hf_model_forward(
                    hf_model, copy.deepcopy(inputs)
                )
                cosmos_hf_logits = test_cosmos_hf_model(
                    cosmos_hf_model, copy.deepcopy(inputs)
                )
                assert torch.equal(
                    hf_generate_logits, hf_forward_logits
                ), f"{hf_generate_logits} != {hf_forward_logits}"
                assert torch.equal(
                    hf_generate_logits, cosmos_hf_logits
                ), f"{hf_generate_logits} != {cosmos_hf_logits}"

                print(f"{model_id} forward test passed.")
                del cosmos_hf_model
                del hf_model
                del cosmos_hf_logits
                del hf_generate_logits
                del hf_forward_logits
                torch.cuda.empty_cache()
            except Exception as e:
                print(f"{model_id} forward test failed.")
                print(f"{e}")
                exit(-1)


if __name__ == "__main__":
    unittest.main()



import os
import unittest
import subprocess
import tempfile

config_content = """
redis = "12808"

[train]
resume = false
epoch = 1
output_dir = "./outputs/qwen2-5-3b-p-fsdp1-tp1-r-tp1-pp1-grpo"
epsilon = 1e-6
optm_name = "AdamW"
optm_lr = 1e-6
optm_impl = "fused"
optm_weight_decay = 0.01
optm_betas = [ 0.9, 0.999,]
optm_warmup_steps = 20
optm_grad_norm_clip = 1.0
async_tp_enabled = false
compile = false
param_dtype = "bfloat16"
fsdp_reduce_dtype = "float32"
fsdp_offload = false
fsdp_reshard_after_forward = "default"
train_batch_per_replica = 32
sync_weight_interval = 1

[rollout]
gpu_memory_utilization = 0.7
enable_chunked_prefill = false
max_response_length = 2048
n_generation = 16
batch_size = 4
quantization = "none"


[policy]
model_name_or_path = "Qwen/Qwen2.5-3B-Instruct"
model_max_length = 4096
model_gradient_checkpointing = false

[logging]
logger = ['console', 'wandb']
project_name = "cosmos_rl"
experiment_name = "None"

[train.train_policy]
type = "grpo"
dataset.name = "JiaxinTsao/math_examples"
prompt_column_name = "prompt"
response_column_name = "result"
dataset.split = "train"
reward_function = "boxed_math"
temperature = 0.9
epsilon_low = 0.2
epsilon_high = 0.2
kl_beta = 0.0
mu_iterations = 1
min_filter_prefix_tokens = 1

[train.ckpt]
enable_checkpoint = false
save_freq = 20
save_mode = "async"

[rollout.parallelism]
n_init_replicas = 1
tp_size = 1
pp_size = 1

[policy.parallelism]
n_init_replicas = 1
tp_size = 1
cp_size = 1
dp_shard_size = 1
pp_size = 1
dp_replicate_size = 1

"""


class TestCustomArgs(unittest.TestCase):
    def test_custom_args(self):
        # save the config file to a temp file
        with tempfile.NamedTemporaryFile(
            mode="w+", suffix=".toml", delete=False
        ) as tmpfile:
            tmpfile.write(config_content)
            tmpfile_toml = tmpfile.name

        current_dir = os.path.dirname(os.path.abspath(__file__))
        config_file = f"{tmpfile_toml}"
        custom_dataset_script = os.path.join(
            current_dir, "custom_dataset/custom_gsm8k_grpo.py"
        )
        cmd = f"cosmos-rl --config {config_file} {custom_dataset_script} --foo cosmos cosmos_rl"
        process = subprocess.Popen(cmd, shell=True)

        process.wait()
        self.assertEqual(process.returncode, 0)


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import List

from cosmos_rl.rollout.schema import RolloutResult

from cosmos_rl.policy.config import Config as CosmosConfig
from transformers import AutoTokenizer


class RolloutBase(ABC):
    def __init__(self, config: CosmosConfig, tokenizer: AutoTokenizer):
        self.config = config
        self.tokenizer = tokenizer
        self.pad_token_id = self.tokenizer.pad_token_id

    @abstractmethod
    def rollout_generation(self, prompts, *args, **kwargs) -> List[RolloutResult]:
        """Generate sequences"""
        pass

    @abstractmethod
    def init_engine(self, quantization: str, seed: int, load_format: str):
        """Initialize the engine"""
        pass



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import torch
import inspect
from torch import nn
from safetensors import safe_open
from transformers.utils import quantization_config as transformers_quantization_config
from functools import partial, cached_property
from typing import Tuple, List, Optional, Callable

from transformers import AutoConfig, AutoModel
from cosmos_rl.utils.util import (
    clear_weight_name,
    safe_deep_getattr,
    load_model_class_by_config,
    reverse_hf_checkpoint_mapping,
    resolve_model_path,
)
from cosmos_rl.utils.constant import COSMOS_HF_MODEL_TYPES
from cosmos_rl.policy.model.base import BaseModel, ModelRegistry
from cosmos_rl.utils.logging import logger
from cosmos_rl.policy.model.hf_models.weight_converter import convert_weight_from_hf
from cosmos_rl.policy.model.hf_models.weight_mapper import HFModelWeightMapper
from cosmos_rl.utils.parallelism import ParallelDims
from cosmos_rl.policy.config import Config as CosmosConfig
from cosmos_rl.policy.model.hf_models.patch import (
    pre_hf_models_patch,
    post_hf_models_patch,
)


@ModelRegistry.register(HFModelWeightMapper)
class HFModel(BaseModel):
    """
    HFModel Module

    Args:
        hf_config : Model configuration arguments.
        model: model loaded from hf.

    Attributes:
        hf_config : Model configuration arguments.
        model: model loaded from hf.
    """

    @staticmethod
    def supported_model_types():
        return [COSMOS_HF_MODEL_TYPES]

    def __init__(
        self, hf_config, model, model_class, is_vlm=False, need_dequantization=False
    ):
        super().__init__(hf_config)
        self.hf_config = hf_config
        self.model = model
        self.model = self.model.to(dtype=hf_config.torch_dtype)
        self.model_class = model_class
        self.is_vlm = is_vlm
        self.need_dequantization = need_dequantization
        if getattr(model, "_checkpoint_conversion_mapping", None):
            if hf_config.model_type in ["R"]:
                logger.warning(
                    f"{hf_config.model_type}'s checkpoint_conversion_mapping do not take effect, "
                    "skip reverse_hf_conversion_mapping"
                )
            else:
                # Reverse HuggingFace checkpoint conversion mapping to align with VLLM weight naming convention
                self.weight_mapper.reverse_hf_conversion_mapping = (
                    reverse_hf_checkpoint_mapping(model._checkpoint_conversion_mapping)
                )
                logger.info(
                    f"reverse_hf_conversion_mapping={self.weight_mapper.reverse_hf_conversion_mapping}"
                )

    @cached_property
    def model_forward_valid_kwargs(self):
        sig = inspect.signature(self.model.forward)
        return sig.parameters.keys()

    def forward(
        self,
        input_ids: torch.Tensor,
        position_ids: torch.Tensor = None,
        *args,
        **kwargs,
    ):
        kwargs_filtered = {
            k: v for k, v in kwargs.items() if k in self.model_forward_valid_kwargs
        }

        out = self.model(
            input_ids=input_ids,
            position_ids=position_ids,
            past_key_values=None,
            use_cache=False,
            *args,
            **kwargs_filtered,
        )
        return out.logits

    @property
    def image_token_id(self):
        # Retrieve image token ID from either image_token_id or image_token_index attribute
        image_token_id = None
        if self.is_vlm:
            image_token_id = getattr(self.hf_config, "image_token_id", None) or getattr(
                self.vision_config, "image_token_id", None
            )
            if image_token_id is None:
                image_token_id = getattr(
                    self.hf_config, "image_token_index", None
                ) or getattr(self.vision_config, "image_token_index", None)
            if image_token_id is None:
                raise ValueError(f"Can not get image token id from {self.hf_config}")
        return image_token_id

    @property
    def video_token_id(self):
        video_token_id = None
        if self.is_vlm:
            video_token_id = getattr(self.hf_config, "video_token_id", None) or getattr(
                self.vision_config, "video_token_id", None
            )
            if video_token_id is None:
                video_token_id = getattr(
                    self.hf_config, "video_token_index", None
                ) or getattr(self.vision_config, "video_token_index", None)
        return video_token_id

    @property
    def lm_layers(self):
        lm_layers = None
        sub_lm_model = getattr(self.language_model, "model", None)
        if sub_lm_model is None:
            lm_layers = self.language_model.layers
        else:
            lm_layers = sub_lm_model.layers
        assert (
            lm_layers is not None
        ), f"Can not get lm layers from {self.language_model}"
        return lm_layers

    @property
    def vision_layers(self):
        vision_layers = None
        if self.vision_model is not None:
            for path in [
                "blocks",  # ClipVisionModel(Llava)
                "vision_model.encoder.layers",  # SiglipVisionModel(Gemma)
                "transformer.layers",  # PixtralVisionModel（Mistral）
                "model.layers",  # Llama4VisionModel
                "encoder.layer",  # InternVLVisionModel(qwen)
                "encoder.layers",  # InternVLVisionModel(gpt-oss)
            ]:
                vision_layers = safe_deep_getattr(self.vision_model, path, None)
                if vision_layers is not None:
                    break
            assert (
                vision_layers is not None
            ), f"Can not get vision layers from {self.vision_model}."
        return vision_layers

    @property
    def n_lm_layers(self):
        n_lm_layers = 0
        if hasattr(self.text_config, "num_hidden_layers"):
            n_lm_layers = self.text_config.num_hidden_layers
        else:
            logger.warning(f"Can not get num of llm layers from {self.text_config}.")
        return n_lm_layers

    @property
    def n_vision_layers(self):
        n_vision_layers = 0
        if hasattr(self.vision_config, "num_hidden_layers"):
            n_vision_layers = self.vision_config.num_hidden_layers
        # qwen2.5-vl
        elif hasattr(self.vision_config, "depth"):
            n_vision_layers = self.vision_config.depth
        else:
            logger.warning(
                f"Can not get num of vision model layers from {self.vision_config}."
            )
        return n_vision_layers

    @property
    def text_config(self):
        text_config = None
        if self.is_vlm:
            if hasattr(self.hf_config, "text_config"):
                text_config = self.hf_config.text_config
            elif hasattr(self.hf_config, "llm_config"):
                text_config = self.hf_config.llm_config
            else:
                logger.warning(f"Can not get text config from {self.hf_config}.")
                text_config = self.hf_config
        else:
            text_config = self.hf_config
        return text_config

    @property
    def vision_config(self):
        return self.hf_config.vision_config if self.is_vlm else None

    @property
    def language_model(self):
        language_model = None
        if self.is_vlm:
            if hasattr(self.model, "language_model"):
                language_model = self.model.language_model
            elif hasattr(self.model, "model"):
                language_model = self.model.model
            else:
                logger.warning(f"Can not get language model from {self.model}.")
        else:
            language_model = self.model
        return language_model

    @property
    def embed_tokens(self):
        embed_tokens = getattr(self.language_model, "embed_tokens", None)
        if embed_tokens is None:
            embed_tokens = safe_deep_getattr(
                self.language_model, "model.embed_tokens", None
            )
        if embed_tokens is None:
            raise ValueError(f"Can not get embed tokens from {self.language_model}")
        return embed_tokens

    @property
    def vision_model(self):
        vision_model = None
        if self.is_vlm:
            # Extract vision model from various possible attribute names
            if hasattr(self.model, "vision_tower"):
                vision_model = self.model.vision_tower
            elif hasattr(self.model, "visual"):
                vision_model = self.model.visual
            elif hasattr(self.model, "vision_model"):
                vision_model = self.model.vision_model
            else:
                raise ValueError(f"Can not get vision model from {self.model}")
        return vision_model

    @property
    def multi_modal_projector(self):
        multi_modal_projector = None
        if self.is_vlm:
            if hasattr(self.model, "multi_modal_projector"):
                multi_modal_projector = self.model.multi_modal_projector
            elif hasattr(self.model, "mlp1"):
                # Handle InternVL architecture's multi-modal projector naming
                multi_modal_projector = self.model.mlp1

        return multi_modal_projector

    @property
    def delay_cp_slice_inputs(self):
        return self.is_vlm

    def post_to_empty_hook(self, cosmos_config: CosmosConfig):
        # Named buffers will be reset during the load_hf_weights process
        return

    @property
    def parallelize_fn(self):
        from cosmos_rl.policy.model.hf_models.parallelize import parallelize

        return parallelize, self

    def apply_pipeline_split(self, pp_rank, pp_size):
        """
        Apply pipeline split to the model.
        This typically involves splitting the model into multiple stages,
        and moving each stage to a different device.
        """
        assert False, "Pipeline split is not supported for HFModel"

    def reset_named_buffers(self, hf_model=None, model_name_or_path=None):
        reset_success = False
        if hf_model is not None:
            # copy named buffers from hf_model to self.model
            hf_named_buffers = {k: v for k, v in hf_model.named_buffers()}
            for name, cosmos_hf_buffer in self.model.named_buffers():
                assert name in hf_named_buffers, f"Buffer {name} not found in hf model"
                hf_buf = hf_named_buffers[name].to(
                    device=cosmos_hf_buffer.device, dtype=cosmos_hf_buffer.dtype
                )
                cosmos_hf_buffer.data.copy_(hf_buf.data)
            reset_success = True
        else:
            assert (
                model_name_or_path is not None
            ), "model_name_or_path is required for resetting named buffers"
            config = AutoConfig.from_pretrained(
                model_name_or_path, trust_remote_code=True
            )

            # Load only first 2 layers instead of full model to extract named buffers efficiently.
            # Most buffers (e.g., inv_freq) are initialized in the model constructor, making this approach sufficient.
            num_lm_layers_to_load = 2
            if self.is_vlm:
                if hasattr(config, "text_config") and hasattr(
                    config.text_config, "num_hidden_layers"
                ):
                    config.text_config.num_hidden_layers = num_lm_layers_to_load
                    config.text_config.max_position_embeddings = (
                        self.hf_config.max_position_embeddings
                    )
                elif hasattr(config, "llm_config") and hasattr(
                    config.llm_config, "num_hidden_layers"
                ):
                    config.llm_config.num_hidden_layers = num_lm_layers_to_load
                    config.llm_config.max_position_embeddings = (
                        self.hf_config.max_position_embeddings
                    )
                else:
                    raise ValueError(f"Can not get text config from {config}")
            else:
                if hasattr(config, "num_hidden_layers"):
                    config.num_hidden_layers = num_lm_layers_to_load
                    config.max_position_embeddings = (
                        self.hf_config.max_position_embeddings
                    )
                else:
                    raise ValueError(f"Can not get num of llm layers from {config}")
            # Attempt to load partial model to extract all named buffers
            try:
                if isinstance(self.model_class, AutoModel):
                    hf_model = AutoModel.from_config(config)
                else:
                    hf_model = self.model_class._from_config(config)
                hf_named_buffers = [name for name, _ in hf_model.named_buffers()]
                self_named_buffers = [name for name, _ in self.model.named_buffers()]
                num_equal = len(hf_named_buffers) == len(self_named_buffers)
                if not num_equal:
                    # Check if the buffers are registered in the layers
                    is_buffer_registered_in_layers = any(
                        "layers." in name for name in hf_named_buffers
                    )
                    if (self.n_lm_layers - num_lm_layers_to_load) == (
                        len(self_named_buffers) - len(hf_named_buffers)
                    ) and is_buffer_registered_in_layers:
                        hf_buffer_in_layers = [
                            buffer
                            for name, buffer in hf_model.named_buffers()
                            if "layers." in name
                        ]
                        first_buffer = hf_buffer_in_layers[0]
                        all_same = True
                        # Verify that all layer buffers contain identical values
                        for buffer in hf_buffer_in_layers[1:]:
                            if not torch.equal(
                                buffer,
                                first_buffer.to(
                                    device=buffer.device, dtype=buffer.dtype
                                ),
                            ):
                                all_same = False
                                break
                        # If all buffers in the layers are the same, we can repeat the first layer's
                        # buffer to the rest of the layers
                        if all_same:
                            cosmos_buffer_in_layers = [
                                buffer
                                for name, buffer in self.model.named_buffers()
                                if "layers." in name
                            ]
                            hf_first_layer_buffer = first_buffer.to(
                                device=cosmos_buffer_in_layers[0].device,
                                dtype=cosmos_buffer_in_layers[0].dtype,
                            )
                            for buffer in cosmos_buffer_in_layers:
                                buffer.data.copy_(hf_first_layer_buffer.data)

                            hf_named_buffers_not_in_layers = {
                                k: v
                                for k, v in hf_model.named_buffers()
                                if "layers." not in k
                            }
                            for name, cosmos_hf_buffer in self.model.named_buffers():
                                if "layers." in name:
                                    continue
                                assert (
                                    name in hf_named_buffers_not_in_layers
                                ), f"Buffer {name} not found in hf model"
                                hf_buf = hf_named_buffers_not_in_layers[name].to(
                                    device=cosmos_hf_buffer.device,
                                    dtype=cosmos_hf_buffer.dtype,
                                )
                                cosmos_hf_buffer.data.copy_(hf_buf.data)

                            return True
                        else:
                            logger.warning(
                                f"Failed to reset named buffers from {model_name_or_path}: buffer names mismatch {self_named_buffers} != {hf_named_buffers}"
                            )
                    else:
                        logger.warning(
                            f"Failed to reset named buffers from {model_name_or_path}: num of buffers mismatch {len(self_named_buffers)} != {len(hf_named_buffers)}"
                        )
                        return False

                reset_success = self.reset_named_buffers(hf_model=hf_model)
            except Exception as e:
                logger.error(
                    f"Failed to reset named buffers from {model_name_or_path}: {e}"
                )
                reset_success = False

        return reset_success

    def load_hf_weights_from_safetensors(
        self,
        model_name_or_path: str,
        parallel_dims: ParallelDims,
        device: torch.device,
        revision: Optional[str] = None,
    ):
        model_type = self.hf_config.model_type
        model_path = resolve_model_path(model_name_or_path, revision=revision)
        safetensors_files = [
            f for f in os.listdir(model_path) if f.endswith(".safetensors")
        ]

        self_state_dict = self.model.state_dict()
        self_state_dict = {clear_weight_name(k): v for k, v in self_state_dict.items()}
        lm_head_weight_key = None
        embed_tokens_weight_key = None
        # Find the lm_head and embed_tokens weight keys in the state dict
        for k in self_state_dict.keys():
            if "embed_tokens" in k:
                embed_tokens_weight_key = k
                if lm_head_weight_key is not None:
                    break
            if "lm_head" in k:
                lm_head_weight_key = k
                if embed_tokens_weight_key is not None:
                    break
        assert (
            lm_head_weight_key is not None and embed_tokens_weight_key is not None
        ), "lm_head and embed_tokens weight keys not found in the state dict"
        weights_of_ckpt_names = set()
        reserved = {}
        hf_checkpoint_conversion_mapping = getattr(
            self.model, "_checkpoint_conversion_mapping", None
        )
        for f in safetensors_files:
            weights_of_ckpt = {}
            ckpt = safe_open(
                os.path.join(model_path, f), framework="pt", device=str(device)
            )
            keys = ckpt.keys()
            for name in keys:
                ckpt_tensor = ckpt.get_tensor(name)
                if hf_checkpoint_conversion_mapping is not None:
                    for (
                        pattern,
                        replacement,
                    ) in hf_checkpoint_conversion_mapping.items():
                        if re.match(pattern, name):
                            name = re.sub(pattern, replacement, name)
                            break
                weights_of_ckpt[name] = ckpt_tensor
                weights_of_ckpt_names.add(name)
                if name == embed_tokens_weight_key:
                    reserved[name] = ckpt_tensor

            for name in weights_of_ckpt.keys():
                tensor = weights_of_ckpt[name]
                dest_name, shared_weight = convert_weight_from_hf(
                    tensor, name, model_type, parallel_dims
                )

                target_tensor = self_state_dict[dest_name]
                is_dist_tensor = isinstance(
                    target_tensor, torch.distributed.tensor.DTensor
                )
                local_view = (
                    target_tensor.to_local() if is_dist_tensor else target_tensor
                )
                assert (
                    local_view.shape == shared_weight.shape
                ), f"Shape mismatch: {local_view.shape} != {shared_weight.shape} for {dest_name}"
                with torch.no_grad():
                    local_view.data.copy_(shared_weight)

        if (
            lm_head_weight_key not in weights_of_ckpt_names
            and embed_tokens_weight_key in weights_of_ckpt_names
        ):
            # Handle weight tying: lm_head shares weights with embed_tokens
            name = lm_head_weight_key
            assert embed_tokens_weight_key in reserved
            tensor = reserved[embed_tokens_weight_key]
            dest_name, shared_weight = convert_weight_from_hf(
                tensor, name, model_type, parallel_dims
            )
            if dest_name in self_state_dict:
                target_tensor = self_state_dict[dest_name]
                is_dist_tensor = isinstance(
                    target_tensor, torch.distributed.tensor.DTensor
                )
                local_view = (
                    target_tensor.to_local() if is_dist_tensor else target_tensor
                )
                assert (
                    local_view.shape == shared_weight.shape
                ), f"Shape mismatch: {local_view.shape} != {shared_weight.shape} for {dest_name}"
                with torch.no_grad():
                    local_view.data.copy_(shared_weight.to(device))

    def load_hf_weights(
        self,
        model_name_or_path: str,
        parallel_dims: ParallelDims,
        device: torch.device,
        revision: Optional[str] = None,
    ):
        """
        Load weights from a HuggingFace model.

        Args:
            model_path (str): Path to the HuggingFace model.
            parallel_dims (ParallelDims): Parallel dimensions definition.
        """
        model_type = self.hf_config.model_type
        dtype = self.hf_config.torch_dtype
        kwargs = {
            "config": self.hf_config,
            "revision": revision,
            "trust_remote_code": True,
        }
        if self.need_dequantization:
            quantization_config = self.hf_config.quantization_config
            mxfp4_quantization_config = transformers_quantization_config.Mxfp4Config(
                dequantize=True,
                modules_to_not_convert=quantization_config["modules_to_not_convert"],
            )
            kwargs["quantization_config"] = mxfp4_quantization_config

        # Configure gradient checkpointing if enabled
        if self._gradient_checkpointing_enabled:
            self.model.gradient_checkpointing_enable()
            assert (
                self.model.is_gradient_checkpointing
            ), "Gradient checkpointing is not enabled"
            logger.info("Enabled gradient checkpointing for HFModel")

        # Use from_pretrained loading in two scenarios:
        # 1. Model requires dequantization (e.g., gpt-oss)
        # 2. Named buffer reinitialization failed
        load_hf_weights_from_pretrained = (
            self.need_dequantization
            or not self.reset_named_buffers(model_name_or_path=model_name_or_path)
        )

        if not load_hf_weights_from_pretrained:
            return self.load_hf_weights_from_safetensors(
                model_name_or_path,
                parallel_dims,
                device,
                revision,
            )

        logger.warning(
            "Loading weights via from_pretrained method - this may take considerable time."
        )

        hf_model = self.model_class.from_pretrained(
            model_name_or_path,
            **kwargs,
        ).to(device="cpu", dtype=dtype)

        self.reset_named_buffers(hf_model=hf_model)

        hf_state_dict = hf_model.state_dict()

        self_state_dict = self.model.state_dict()
        self_state_dict = {clear_weight_name(k): v for k, v in self_state_dict.items()}

        for name, tensor in hf_state_dict.items():
            dest_name, shared_weight = convert_weight_from_hf(
                tensor, name, model_type, parallel_dims
            )

            target_tensor = self_state_dict[dest_name]
            is_dist_tensor = isinstance(target_tensor, torch.distributed.tensor.DTensor)
            local_view = target_tensor.to_local() if is_dist_tensor else target_tensor
            assert (
                local_view.shape == shared_weight.shape
            ), f"Shape mismatch: {local_view.shape} != {shared_weight.shape} for {dest_name}"
            with torch.no_grad():
                local_view.data.copy_(shared_weight.to(device))

        del hf_model

    def get_position_ids(self, **kwargs) -> Tuple[torch.Tensor, torch.Tensor, int]:
        position_ids = None
        inputs = kwargs["input_ids"]
        seq_dim_idx = 1 if not self.is_vlm else 2
        return position_ids, inputs, seq_dim_idx

    def separate_model_parts(self) -> List[nn.Module]:
        if self.is_vlm:
            model_parts = [self.language_model, self.vision_model]
            if self.multi_modal_projector is not None:
                logger.info("Adding multi_modal_projector to model parts")
                model_parts.append(self.multi_modal_projector)
            # Handle cases where lm_head exists at model level rather than language_model level
            if (
                getattr(self.language_model, "lm_head", None) is None
                and getattr(self.model, "lm_head", None) is not None
            ):
                logger.info("Adding lm_head to model parts")
                model_parts.append(self.model.lm_head)
            return model_parts
        else:
            return [self.language_model]

    @cached_property
    def _get_nparams_and_flops_fn(
        self, is_vision_model: bool
    ) -> Callable[[int], tuple[int, int]]:
        if is_vision_model:
            nparams = sum(p.numel() for p in self.parameters())
            nparams_embedding = sum(
                sum(p.numel() for p in m.parameters())
                for m in self.children()
                if isinstance(m, nn.Embedding)
            )

            # Reasoning behind the factor of 12 for the self-attention part of the formula:
            # 1. each self-attention has 2 matmul in the forward and 4 in the backward (6)
            # 2. the flash attention does 1 more matmul recomputation in the backward
            #    but recomputation should not be counted in calculating MFU           (+0)
            # 3. each matmul performs 1 multiplication and 1 addition                 (*2)
            # 4. we follow the convention and do not account for sparsity in causal attention
            num_heads = 0
            if hasattr(self.vision_config, "num_attention_heads"):
                num_heads = self.vision_config.num_attention_heads
            elif hasattr(self.vision_config, "num_heads"):
                num_heads = self.vision_config.num_heads
            else:
                logger.warning(f"Can not get num of heads from {self.vision_config}.")
            layers, heads, head_dim = (
                self.n_vision_layers,
                num_heads,
                self.vision_config.hidden_size // num_heads,
            )
            return lambda seq_len: (
                nparams,
                6 * (nparams - nparams_embedding)
                + 12 * layers * heads * head_dim * seq_len,
            )
        else:
            nparams = sum(p.numel() for p in self.parameters())
            nparams_embedding = sum(
                sum(p.numel() for p in m.parameters())
                for m in self.children()
                if isinstance(m, nn.Embedding)
            )
            # Reasoning behind the factor of 12 for the self-attention part of the formula:
            # 1. each self-attention has 2 matmul in the forward and 4 in the backward (6)
            # 2. the flash attention does 1 more matmul recomputation in the backward
            #    but recomputation should not be counted in calculating MFU           (+0)
            # 3. each matmul performs 1 multiplication and 1 addition                 (*2)
            # 4. we follow the convention and do not account for sparsity in causal attention
            layers, heads, head_dim = (
                self.n_lm_layers,
                self.text_config.num_attention_heads,
                self.text_config.hidden_size // self.text_config.num_attention_heads,
            )
            return lambda seq_len: (
                nparams,
                6 * (nparams - nparams_embedding)
                + 12 * layers * heads * head_dim * seq_len,
            )

    def get_nparams_and_flops(self, seq_len: int) -> tuple[int, int]:
        n_params = 0
        n_flops = 0
        if self.vision_model is not None:
            n_params, n_flops = self._get_nparams_and_flops_fn(is_vision_model=True)(
                seq_len
            )

        lm_n_params, lm_n_flops = self._get_nparams_and_flops_fn(is_vision_model=False)(
            seq_len
        )
        n_params += lm_n_params
        n_flops += lm_n_flops
        return n_params, n_flops

    @classmethod
    def from_model_args(cls, hf_config) -> "HFModel":
        """
        Initialize a HFModel model from a HFModelArgs object.

        Args:
            hf_config : hf model config.

        Returns:
            HFModel: HFModel model.

        """
        is_vlm = getattr(hf_config, "vision_config", None) is not None
        model_class = None
        quantization_config = getattr(hf_config, "quantization_config", None)
        need_dequantization = False
        if quantization_config is not None:
            if quantization_config["quant_method"] in ["mxfp4"]:
                assert hasattr(
                    transformers_quantization_config, "Mxfp4Config"
                ), "Mxfp4Config is not supported in this version of transformers. Please upgrade transformers to version 4.45.0 or higher."
                logger.warning(
                    "We don't support mxfp4 training for HFModel currently, will default to dequantizing the model to bf16/fp16."
                )
                need_dequantization = True
            hf_config.quantization_config["dequantize"] = need_dequantization

        pre_hf_models_patch(hf_config)

        try:
            model_class = load_model_class_by_config(hf_config)
            model = model_class(hf_config)
        except Exception as e:
            logger.warning(
                f"Got error({e}) when loading {hf_config.model_type}, Using AutoModel instead."
            )
            model_class = AutoModel
            model = AutoModel.from_config(hf_config, trust_remote_code=True)

        post_hf_models_patch(hf_config, model)

        return cls(
            hf_config,
            model,
            model_class,
            is_vlm=is_vlm,
            need_dequantization=need_dequantization,
        )

    @classmethod
    def from_pretrained(
        cls,
        hf_config: AutoConfig,
        model_name_or_path: str,
        max_position_embeddings: Optional[int] = None,
    ) -> "HFModel":
        """
        Initialize a HFModel model from a pretrained model.

        Args:
            hf_config (AutoConfig): HuggingFace config.
            model_name_or_path (str): Model name or path to the pretrained model.
            max_position_embeddings (int): Maximum position embeddings.

        Returns:
            HFModel: HFModel model.

        """

        if max_position_embeddings is not None:
            hf_config.max_position_embeddings = max_position_embeddings
            if hasattr(hf_config, "text_config") and hasattr(
                hf_config.text_config, "max_position_embeddings"
            ):
                hf_config.text_config.max_position_embeddings = max_position_embeddings
            elif hasattr(hf_config, "llm_config") and hasattr(
                hf_config.llm_config, "max_position_embeddings"
            ):
                hf_config.llm_config.max_position_embeddings = max_position_embeddings

        return cls.from_model_args(hf_config)

    def named_parameters(self, *args, **kwargs):
        return self.model.named_parameters(*args, **kwargs)

    @classmethod
    def fqn_filter_for_fp8(cls) -> List[str]:
        llm = [
            "lm_head",
        ]
        visual = [
            "visual",
            "vision_tower",
        ]  # Filter Linear in visual out, they will corrupt the FP8 Linear.
        return llm + visual

    def check_cp_compatible(self, cp_size: int, tp_size: int):
        assert cp_size == 1, "cp is not supported for HFModel"
        assert tp_size == 1, "tp is not supported for HFModel"

    def post_transform_of_local_view(self, local_view: torch.Tensor, name: str):
        if "gpt_oss" in self.hf_config.model_type:
            if "bias" not in name:  # Bias parameters do not require transposition
                if "gate_up_proj" in name or "down_proj" in name:

                    def transform(view):
                        return view.transpose(-2, -1).contiguous()

                    return partial(transform, local_view)

        return local_view



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
API client for the dispatcher.
"""

import os
import re
import requests
import msgpack
from functools import partial
from typing import Dict, Any, List, Tuple, Optional
from urllib.parse import urljoin

from cosmos_rl.dispatcher.protocol import (
    Role,
    ValidationReportRequest,
    RolloutRequest,
    SetProfileRequest,
    SetTracePathRequest,
)
from cosmos_rl.utils.network_util import make_request_with_retry
from cosmos_rl.utils import constant
from cosmos_rl.utils.api_suffix import (
    COSMOS_API_STATUS_SUFFIX,
    COSMOS_API_META_SUFFIX,
    COSMOS_API_REGISTER_SUFFIX,
    COSMOS_API_SET_PROFILE_SUFFIX,
    COSMOS_API_SET_TRACE_PATH_SUFFIX,
    COSMOS_API_UNREGISTER_SUFFIX,
    COSMOS_API_HEARTBEAT_SUFFIX,
    COSMOS_API_NCCL_COMM_INITIATOR_SUFFIX,
    COSMOS_API_NCCL_COMM_ACCEPTOR_SUFFIX,
    COSMOS_API_NCCL_COMM_GET_ALL_SUFFIX,
    COSMOS_API_NCCL_COMM_ERROR_SUFFIX,
    COSMOS_API_NCCL_COMM_STORE_CLEAR_SUFFIX,
    COSMOS_API_NEXT_PROMPT_SUFFIX,
    COSMOS_API_ROLLOUT_SUFFIX,
    COSMOS_API_VALIDATION_REPORT_SUFFIX,
    COSMOS_API_POLICY_TRAIN_ACK_SUFFIX,
    COSMOS_API_POLICY_SHARD_INFOS_SUFFIX,
    COSMOS_API_ROLLOUT_SHARD_INFOS_SUFFIX,
    COSMOS_API_POLICY_SHARD_SEND_INSTS_SUFFIX,
    COSMOS_API_ROLLOUT_SHARD_RECV_INSTS_SUFFIX,
    COSMOS_API_GET_TRAINABLE_PARAMS_SUFFIX,
)
from cosmos_rl.utils.parallelism_map import WeightSyncInstructionsGroup
from cosmos_rl.utils.util import list_to_b64, sanitize, b64_to_list
from cosmos_rl.utils.logging import logger


class APIClient(object):
    """
    API client for the dispatcher.
    """

    def __init__(
        self,
        role: Role,
        remote_ips: Optional[List[str]] = None,
        remote_port: Optional[int] = None,
    ):
        self.role = role

        self.remote_ips = remote_ips
        self.remote_port = remote_port
        self.base_urls = []
        self.__update_base_urls()

        self.max_retries = constant.COSMOS_HTTP_RETRY_CONFIG.max_retries

    def __update_base_urls(self):
        if self.remote_ips is None or self.remote_port is None:
            # parser remote_hosts from env COSMOS_CONTROLLER_HOST
            remote_hosts = os.environ["COSMOS_CONTROLLER_HOST"]
            remote_ips, remote_port = remote_hosts.split(":")
            remote_ips = remote_ips.split(";")
            self.remote_ips = remote_ips
            self.remote_port = remote_port

        # Verify in the format of host:port
        for remote_ip in self.remote_ips:
            if not re.match(
                r"^([a-zA-Z0-9_.-]+):([1-9][0-9]{0,4})$",
                f"{remote_ip}:{self.remote_port}",
            ):
                raise ValueError(f"Invalid remote host: {remote_ip}:{self.remote_port}")

        self.base_urls = [
            f"http://{remote_ip}:{self.remote_port}{COSMOS_API_META_SUFFIX}"
            for remote_ip in self.remote_ips
        ]

    def get_alternative_urls(self, suffix: str):
        urls = []
        for base_url in self.base_urls:
            urls.append(urljoin(base_url, suffix))
        return urls

    def get_controller_metadata(self) -> Dict[str, Any]:
        """
        Get metadata from the controller with retry logic.
        """
        try:
            r = make_request_with_retry(
                partial(
                    requests.get,
                ),
                self.get_alternative_urls(COSMOS_API_META_SUFFIX),
                max_retries=self.max_retries,
            )
            metadata = r.json()
        except Exception as e:
            logger.error(f"Failed to communicate with controller after attempts: {e}")
            raise e

        # update base urls from controller
        remote_eth_ips = metadata.get("config", {}).get("eth_ips", [])
        if remote_eth_ips:
            remote_ips = set(self.remote_ips)
            remote_ips.update(remote_eth_ips.split(";"))
            self.remote_ips = list(remote_ips)
        self.__update_base_urls()
        return metadata

    def register(
        self,
        replica_name: str,
        role: Role,
        mesh_names: List[str],
        ranks: List[int],
        group_size: List[int],
        global_rank: int,
        host_ip: str,
        host_name: str,
    ):
        try:
            make_request_with_retry(
                partial(
                    requests.post,
                    json={
                        "replica_name": replica_name,
                        "role": role,
                        "mesh_names": mesh_names,
                        "ranks": ranks,
                        "group_size": group_size,
                        "global_rank": global_rank,
                        "host_ip": host_ip,
                        "host_name": host_name,
                    },
                ),
                self.get_alternative_urls(COSMOS_API_REGISTER_SUFFIX),
                max_retries=self.max_retries,
            )
        except Exception as e:
            logger.error(f"Failed to register to controller: {e}")
            raise e

    def unregister(self, replica_name: str):
        try:
            make_request_with_retry(
                partial(
                    requests.post,
                    json={"replica_name": replica_name},
                ),
                self.get_alternative_urls(COSMOS_API_UNREGISTER_SUFFIX),
                max_retries=self.max_retries,
            )
        except Exception as e:
            logger.error(f"Failed to unregister from controller: {e}")

    def post_heartbeat(self, replica_name: str):
        try:
            make_request_with_retry(
                partial(
                    requests.post,
                    json={"replica_name": replica_name},
                ),
                self.get_alternative_urls(COSMOS_API_HEARTBEAT_SUFFIX),
                max_retries=self.max_retries,
            )
        except Exception as e:
            logger.error(f"Failed to send heartbeat to controller: {e}")

    def get_status(self) -> Dict[str, Any]:
        try:
            r = make_request_with_retry(
                partial(
                    requests.get,
                ),
                self.get_alternative_urls(COSMOS_API_STATUS_SUFFIX),
                max_retries=self.max_retries,
            )
            return r.json()
        except Exception as e:
            logger.error(f"Failed to get status from controller: {e}")
            raise e

    def post_profile(self, profile: SetProfileRequest):
        try:
            make_request_with_retry(
                partial(
                    requests.post,
                    json=profile.model_dump(),
                ),
                self.get_alternative_urls(COSMOS_API_SET_PROFILE_SUFFIX),
                max_retries=self.max_retries,
            )
        except Exception as e:
            logger.error(f"Failed to set profile to controller: {e}")
            raise e

    def post_trace_path(self, trace_path: SetTracePathRequest):
        try:
            make_request_with_retry(
                partial(
                    requests.post,
                    json=trace_path.model_dump(),
                ),
                self.get_alternative_urls(COSMOS_API_SET_TRACE_PATH_SUFFIX),
                max_retries=self.max_retries,
            )
        except Exception as e:
            logger.error(f"Failed to set trace path to controller: {e}")
            raise e

    def post_nccl_comm_initiator(self, unique_pair_name: str, nccl_uuid: List[int]):
        base64_nccl_group_id = list_to_b64(nccl_uuid)
        try:
            make_request_with_retry(
                partial(
                    requests.post,
                    json={
                        "unique_pair_name": unique_pair_name,
                        "handle_base64": base64_nccl_group_id,
                    },
                ),
                self.get_alternative_urls(COSMOS_API_NCCL_COMM_INITIATOR_SUFFIX),
                max_retries=self.max_retries,
            )
        except Exception as e:
            raise RuntimeError(
                f"[{self.role}] Failed in post nccl group_id to controller after retries {e}."
            )

    def post_nccl_comm_acceptor(self, unique_pair_name: str) -> List[int]:
        try:
            r = make_request_with_retry(
                partial(
                    requests.post,
                    json={"unique_pair_name": unique_pair_name},
                ),
                self.get_alternative_urls(COSMOS_API_NCCL_COMM_ACCEPTOR_SUFFIX),
                max_retries=self.max_retries,
            )
            base64_nccl_group_id = r.json()["handle_base64"]
            return b64_to_list(base64_nccl_group_id)
        except Exception as e:
            raise RuntimeError(
                f"[{self.role}] Failed in get nccl group_id from controller after retries {e}."
            )

    def post_nccl_comm_error(self, replica_name: str, error: Exception):
        try:
            make_request_with_retry(
                partial(
                    requests.post,
                    json={"replica_name": replica_name, "error": str(error)},
                ),
                self.get_alternative_urls(COSMOS_API_NCCL_COMM_ERROR_SUFFIX),
                max_retries=self.max_retries,
            )
        except Exception as e:
            raise RuntimeError(
                f"[{self.role}] Failed in post nccl comm error to controller after retries {e}."
            )

    def post_clear_nccl_comm_store(self, unique_pair_name: str):
        try:
            make_request_with_retry(
                partial(
                    requests.post,
                    json={"unique_pair_name": unique_pair_name},
                ),
                self.get_alternative_urls(COSMOS_API_NCCL_COMM_STORE_CLEAR_SUFFIX),
                max_retries=self.max_retries,
            )
        except Exception as e:
            raise RuntimeError(
                f"[{self.role}] Failed in clear nccl comm store from controller after retries {e}."
            )

    def list_nccl_comm_infos(self) -> List[Dict[str, Any]]:
        """
        List all the NCCL communicators stored in the controller.
        """
        try:
            r = make_request_with_retry(
                partial(
                    requests.get,
                ),
                self.get_alternative_urls(COSMOS_API_NCCL_COMM_GET_ALL_SUFFIX),
                max_retries=self.max_retries,
            )
            comm_info = r.json()["comm_info"]
            comm_dict = {}
            for key, value in comm_info.items():
                comm_dict[key] = str(b64_to_list(value))
            return comm_dict
        except Exception as e:
            raise RuntimeError(
                f"[{self.role}] Failed in list nccl comm from controller after retries {e}."
            )

    def post_policy_shard_info(
        self,
        shard_infos: List[Dict[str, Any]],
        param_groups: List[List[str]],
        sorted_params: List[List[str]],
        trainable_params: List[str],
    ):
        data = msgpack.packb(
            {
                "shard_infos": shard_infos,
                "param_groups": param_groups,
                "sorted_params": sorted_params,
                "trainable_params": trainable_params,
            }
        )
        try:
            make_request_with_retry(
                partial(
                    requests.post,
                    data=data,
                    headers={"Content-Type": "application/msgpack"},
                ),
                self.get_alternative_urls(COSMOS_API_POLICY_SHARD_INFOS_SUFFIX),
                max_retries=self.max_retries,
            )
        except Exception as e:
            raise RuntimeError(
                f"[Policy] Failed to post policy shard infos to controller after retries {e}."
            )

    def post_rollout_shard_info(
        self,
        shard_infos: List[Dict[str, Any]],
        param_groups: List[List[str]],
        sorted_params: List[List[str]],
    ):
        data = msgpack.packb(
            {
                "shard_infos": shard_infos,
                "param_groups": param_groups,
                "sorted_params": sorted_params,
            }
        )
        try:
            make_request_with_retry(
                partial(
                    requests.post,
                    data=data,
                    headers={"Content-Type": "application/msgpack"},
                ),
                self.get_alternative_urls(COSMOS_API_ROLLOUT_SHARD_INFOS_SUFFIX),
                max_retries=self.max_retries,
            )
        except Exception as e:
            raise RuntimeError(
                f"[Rollout] Failed in post shard infos to controller after retries {e}."
            )

    def post_policy_shard_send_insts(
        self, rank: int
    ) -> List[WeightSyncInstructionsGroup]:
        try:
            insts_meta = make_request_with_retry(
                partial(
                    requests.post,
                    json={"rank": rank},
                ),
                self.get_alternative_urls(COSMOS_API_POLICY_SHARD_SEND_INSTS_SUFFIX),
                max_retries=self.max_retries,
            )
            insts = msgpack.unpackb(insts_meta.content, strict_map_key=False)
            return [WeightSyncInstructionsGroup.from_dict(inst) for inst in insts]
        except Exception as e:
            raise RuntimeError(
                f"[Policy] Failed in post policy shard send insts to controller after retries {e}."
            )

    def post_rollout_shard_recv_insts(
        self, rank: int
    ) -> List[WeightSyncInstructionsGroup]:
        try:
            insts_meta = make_request_with_retry(
                partial(
                    requests.post,
                    json={"rank": rank},
                ),
                self.get_alternative_urls(COSMOS_API_ROLLOUT_SHARD_RECV_INSTS_SUFFIX),
                max_retries=self.max_retries,
            )
            insts = msgpack.unpackb(insts_meta.content, strict_map_key=False)
            return [WeightSyncInstructionsGroup.from_dict(inst) for inst in insts]
        except Exception as e:
            raise RuntimeError(
                f"[Rollout] Failed in fetching rollout shard recv insts from controller after retries {e}."
            )

    def post_policy_train_ack(
        self,
        replica_name: str,
        weight_step: int,
        total_steps: int,
        profile_finished: bool,
        report_data: Dict[str, Any],
    ):
        try:
            make_request_with_retry(
                partial(
                    requests.post,
                    json={
                        "replica_name": replica_name,
                        "weight_step": weight_step,
                        "total_steps": total_steps,
                        "profile_finished": profile_finished,
                        "report_data": sanitize(report_data),
                    },
                ),
                self.get_alternative_urls(COSMOS_API_POLICY_TRAIN_ACK_SUFFIX),
                max_retries=self.max_retries,
            )
        except Exception as e:
            raise RuntimeError(
                f"[Policy] Failed in in send train ack to controller after retries {e}."
            )

    def get_trainable_params(self) -> List[str]:
        try:
            r = make_request_with_retry(
                partial(
                    requests.get,
                ),
                self.get_alternative_urls(COSMOS_API_GET_TRAINABLE_PARAMS_SUFFIX),
                max_retries=self.max_retries,
            )
            return r.json()["trainable_params"]
        except Exception as e:
            raise RuntimeError(
                f"[Rollout] Failed in fetching trainable params from controller after retries {e}."
            )

    def post_validation_report(self, report: ValidationReportRequest):
        try:
            make_request_with_retry(
                partial(
                    requests.post,
                    json=report.model_dump(),
                ),
                self.get_alternative_urls(COSMOS_API_VALIDATION_REPORT_SUFFIX),
                max_retries=self.max_retries,
            )
        except Exception as e:
            logger.error(
                f"[Rollout] Failed in sending validation report to controller after retries {e}."
            )

    def get_next_prompt(
        self, batch_size: int, validation_step: Optional[int] = None
    ) -> Tuple[List[Tuple[int, str]], bool]:
        try:
            r = make_request_with_retry(
                partial(
                    requests.get,
                    params={"n": batch_size, "validation_step": validation_step},
                ),
                self.get_alternative_urls(COSMOS_API_NEXT_PROMPT_SUFFIX),
                max_retries=self.max_retries,
            )
            r = r.json()
            payload = r["prompt_id_and_payload_list"]
            is_end = r["is_end"]
            return payload, is_end
        except Exception as e:
            logger.error(
                f"[Rollout] Failed in fetching next prompt from controller after retries {e}."
            )
            return [], False

    def post_rollout_completion(self, response: RolloutRequest):
        try:
            make_request_with_retry(
                partial(
                    requests.post,
                    json=response.model_dump(),
                ),
                self.get_alternative_urls(COSMOS_API_ROLLOUT_SUFFIX),
                max_retries=self.max_retries,
            )
        except Exception as e:
            logger.error(
                f"[Rollout] Failed in sending rollout completion to controller after retries {e}."
            )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Tuple
from cosmos_rl.dispatcher.data.schema import Rollout, RLPayload


def extract_rollouts(
    payloads: List[RLPayload], is_end: bool, prompt_idxs: List[int] = []
) -> Tuple[List[List[Rollout]], List[List[Rollout]]]:
    # Extract rollouts from payloads of request
    # Separate valid and invalid rollouts for Dynamic Sampling
    # Dynamic Sampling: Filter out the rollouts that the rewards are all the same
    valid_rollouts_list: List[List[Rollout]] = []
    invalid_rollouts_list: List[List[Rollout]] = []
    for idx, payload in enumerate(payloads):
        assert (
            len(payload.completions)
            == len(payload.completed_conversations)
            == len(payload.rewards)
            == len(payload.advantages)
            == len(payload.n_ignore_prefix_tokens)
        ), "Length of completions, completed_conversations, rewards, advantages and n_ignore_prefix_tokens must be the same"
        if len(prompt_idxs) != len(payloads):
            prompt_idx = -1
        else:
            prompt_idx = prompt_idxs[idx]
        rollouts = [
            Rollout(
                prompt=payload.prompt,
                conversation=payload.conversation,
                completion=completion,
                completed_conversation=completed_conversation,
                is_end=is_end,
                reward=reward,
                advantage=advantage,
                prompt_idx=prompt_idx,
                n_ignore_prefix_tokens=n_ignore_prefix_tokens,
            )
            for completion, completed_conversation, reward, advantage, n_ignore_prefix_tokens in zip(
                payload.completions,
                payload.completed_conversations,
                payload.rewards,
                payload.advantages,
                payload.n_ignore_prefix_tokens,
            )
        ]
        if payload.valid:
            valid_rollouts_list.append(rollouts)
        else:
            invalid_rollouts_list.append(rollouts)
    return valid_rollouts_list, invalid_rollouts_list



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from contextlib import contextmanager
from inspect import isgeneratorfunction
from typing import Any, Optional

from transformers.utils import get_json_schema

from .schema import OpenAIFunctionToolSchema, ToolResponse


class BaseTool(ABC):
    def __init__(self, name: str, schema: Optional[OpenAIFunctionToolSchema] = None):
        self.name = name
        self._schema = schema
        self._validate_tool_context()

    def __call__(self, *args, **kwargs) -> Any:
        return self.function(*args, **kwargs)

    def _validate_tool_context(self):
        """Make sure the tool_context method is implemented correctly"""
        if not hasattr(self, "tool_context"):
            raise NotImplementedError("tool_context method must be implemented")
        if not hasattr(self.tool_context, "__wrapped__"):
            raise TypeError(
                "tool_context method must be decorated with @contextmanager"
            )
        if not isgeneratorfunction(self.tool_context.__wrapped__):
            raise TypeError("tool_context method must be a generator function")

    @abstractmethod
    @contextmanager
    def tool_context(self, *args, **kwargs):
        """
        This method is optional and get called by launcher after being mounted
        """
        raise NotImplementedError("This method should be implemented by the subclass")

    @property
    def tool_schema(self) -> OpenAIFunctionToolSchema:
        """
        The schema of the tool.
        """
        if self._schema is None:
            # create the default schema for the tool
            _schema = get_json_schema(self.function)
            self._schema = OpenAIFunctionToolSchema.model_validate(_schema)
            self._schema.function.name = self.name
        return self._schema

    @abstractmethod
    def function(self, *args, **kwargs) -> ToolResponse:
        """
        The function to be called when the tool is used.
        """
        raise NotImplementedError("This method should be implemented by the subclass")



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch.distributed as dist
import uuid
from typing import Dict, Callable, Type, Optional, Any
import copy
import time
import atexit
import threading
from cosmos_rl.utils.redis_stream import RedisStreamHandler
from cosmos_rl.utils.network_util import get_local_ip
from cosmos_rl.dispatcher.command import (
    PolicyCommandRegistry,
    RolloutCommandRegistry,
    Command,
)
from cosmos_rl.dispatcher.data.packer import DataPacker, DecoderOnlyLLMDataPacker
from cosmos_rl.dispatcher.data.packer import (
    HFVLMDataPacker,
)

from cosmos_rl.utils.logging import logger
import cosmos_rl.utils.constant as constant
import cosmos_rl.utils.distributed as dist_utils
from cosmos_rl.dispatcher.protocol import MESH_NAMES
import cosmos_rl.utils.util as util
import base64
import cloudpickle
from transformers import AutoConfig
import multiprocessing as mp
from cosmos_rl.dispatcher.api.client import APIClient


class CommMixin:
    policy_command_handler_registry = PolicyCommandRegistry()
    rollout_command_handler_registry = RolloutCommandRegistry()

    @classmethod
    def register_policy_command_handler(cls, command_type: Type[Command]):
        def decorator(func):
            cls.policy_command_handler_registry.register(command_type, func)
            return func

        return decorator

    @classmethod
    def register_rollout_command_handler(
        cls, command_type: Type[Command], backend: str = "vllm"
    ):
        def decorator(func):
            cls.rollout_command_handler_registry.register(command_type, func, backend)
            return func

        return decorator

    @classmethod
    def get_policy_command_handler(
        cls, command_type: Type[Command]
    ) -> Optional[Callable]:
        return cls.policy_command_handler_registry.get_command_handler(command_type)

    @classmethod
    def get_rollout_command_handler(
        cls, command_type: Type[Command], backend: str = "vllm"
    ) -> Optional[Callable]:
        return cls.rollout_command_handler_registry.get_command_handler(
            command_type, backend
        )

    def init_comm(self):
        self.replica_name = str(dist_utils.broadcast_object_cpu(uuid.uuid4()))
        logger.info(
            f"{self.role} Replica started at global rank {self.global_rank}, with replica name: {self.replica_name}"
        )

        self.api_client = APIClient(self.role)
        self.init_meta()

        self.register_to_controller()

    def init_meta(self):
        # Fetch metadata from the controller
        metadata = self.api_client.get_controller_metadata()
        self.init_data_packer(metadata)

    def init_data_packer(self, metadata: Dict[str, Any]):
        hf_config = util.retry(AutoConfig.from_pretrained)(
            self.config.policy.model_name_or_path, trust_remote_code=True
        )
        is_vlm = getattr(hf_config, "vision_config", None) is not None
        model_type = hf_config.model_type

        user_data_packer = metadata.get("user_data_packer", None)
        if user_data_packer:
            user_data_packer = base64.b64decode(user_data_packer)
            user_data_packer = cloudpickle.loads(user_data_packer)
            self.data_packer = user_data_packer
            logger.info(f"Using user-provided data packer: {self.data_packer}")
        else:
            try:
                self.data_packer = DataPacker.get_default_data_packer(model_type)
                logger.info(f"Using default data packer: {self.data_packer}")
            except ValueError:
                self.data_packer = (
                    DecoderOnlyLLMDataPacker() if not is_vlm else HFVLMDataPacker()
                )
                logger.warning(
                    f"No default data packer found for {model_type}, using {type(self.data_packer).__name__} as default"
                )

        self.data_packer.setup(self.config, self.tokenizer)

        user_val_data_packer = metadata.get("user_val_data_packer", None)
        if user_val_data_packer:
            user_val_data_packer = base64.b64decode(user_val_data_packer)
            user_val_data_packer = cloudpickle.loads(user_val_data_packer)
            self.val_data_packer = user_val_data_packer
            logger.info(
                f"Using user-provided validation data packer: {self.val_data_packer}"
            )
        else:
            try:
                self.val_data_packer = DataPacker.get_default_data_packer(model_type)
                logger.info(
                    f"Using default validation data packer: {self.val_data_packer}"
                )
            except ValueError:
                self.val_data_packer = (
                    DecoderOnlyLLMDataPacker() if not is_vlm else HFVLMDataPacker()
                )
                logger.warning(
                    f"No default validation data packer found for {model_type}, using {type(self.val_data_packer).__name__} as default"
                )

        self.val_data_packer.setup(self.config, self.tokenizer)

    def register_to_controller(self):
        if hasattr(self, "_is_registered"):
            return

        target_mesh_names = copy.deepcopy(MESH_NAMES)
        ranks = []
        group_size = []
        for mesh_name in MESH_NAMES:
            if (
                self.parallel_dims.mesh.mesh_dim_names
                and mesh_name in self.parallel_dims.mesh.mesh_dim_names
            ):
                ranks.append(self.parallel_dims.mesh[mesh_name].get_local_rank())
                group_size.append(self.parallel_dims.mesh[mesh_name].size())
            else:
                ranks.append(0)
                group_size.append(1)

        host_info_tuple = get_local_ip()
        if host_info_tuple is None:
            raise Exception("Failed to get local IP address")
        host_ip, host_name = host_info_tuple
        self.api_client.register(
            replica_name=self.replica_name,
            role=self.role,
            mesh_names=target_mesh_names,
            ranks=ranks,
            group_size=group_size,
            global_rank=self.global_rank,
            host_ip=host_ip,
            host_name=host_name,
        )

        dist.barrier()  # wait all the atoms registered.

        self.shutdown_signal = threading.Event()
        self.shutdown_mp_signal = mp.Event()  # Must be a multiprocessing event

        if self.global_rank == 0:
            logger.info(
                f"{self.role} Replica {self.replica_name} registered to controller"
            )
            # Start the thread with daemon=True, so it will exit when the main program exits.
            process = mp.Process(
                target=self.heartbeat_trigger,
                args=(self.shutdown_mp_signal,),
                daemon=True,  # Dies when main process exits
            )
            process.start()
            self.heartbeat_thread = process
        else:
            self.heartbeat_thread = None

        self._is_registered = True
        atexit.register(self.unregister_from_controller)

    def unregister_from_controller(self):
        if not hasattr(self, "_is_registered"):
            return
        elif hasattr(self, "_is_unregistered"):
            return
        else:
            self._is_unregistered = True
        self._is_registered = False
        # let only rank == 0 send the unregister request
        if self.global_rank == 0:
            self.api_client.unregister(self.replica_name)

    def get_group_unique_key(self, replica_name_to_rank: Dict[str, int]):
        return (
            "_".join(
                [
                    k
                    for k, _ in sorted(
                        replica_name_to_rank.items(), key=lambda item: item[1]
                    )
                ]
            )
            + "_"
            + str(self.global_rank)
        )

    def init_redis(self):
        assert (
            self.api_client.remote_ips is not None
        ), "Please init the api client first"
        # For command fetch via redis connection
        self.redis_controller = RedisStreamHandler(
            ips=self.api_client.remote_ips, port=int(self.config.redis)
        )
        logger.debug(
            f"[{self.role}] Init redis at {self.api_client.remote_ips}:{self.redis_controller.port}"
        )

    def heartbeat_trigger(self, shutdown_signal: threading.Event):
        while True:
            self.api_client.post_heartbeat(self.replica_name)

            # If the heartbeat interval is greater than 1, we need to check the shutdown signal every second
            # for faster shutdown check
            if constant.COSMOS_HEARTBEAT_SEND_INTERVAL > 1:
                early_break = False
                for _ in range(int(constant.COSMOS_HEARTBEAT_SEND_INTERVAL)):
                    if shutdown_signal.is_set():
                        early_break = True
                        break
                    else:
                        time.sleep(1)
                if early_break:
                    break
            else:
                time.sleep(constant.COSMOS_HEARTBEAT_SEND_INTERVAL)
                if shutdown_signal.is_set():
                    break



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

# Set the environment variable to use HF rotary implementation
os.environ["COSMOS_USE_HF_IMPL"] = "1"
import torch
import unittest
import numpy as np
from functools import partial

from cosmos_rl.policy.model import ModelRegistry
from cosmos_rl.utils.parallelism import ParallelDims
from cosmos_rl.policy.trainer.sft_trainer import async_safe_ce
from cosmos_rl.dispatcher.data.packer.qwen3_vl_data_packer import (
    Qwen3_VL_DataPacker,
)
from transformers import (
    AutoProcessor,
    AutoTokenizer,
    Qwen3VLMoeForConditionalGeneration,
)


IGNORE_INDEX = -100


def pp_loss_fn(config, parallel_dims):
    loss_scaling_factor = 1.0
    if parallel_dims.dp_shard_enabled:
        dp_group = parallel_dims.mesh["dp_shard"].get_group()
    else:
        dp_group = None

    if parallel_dims.cp_enabled:
        cp_group = parallel_dims.mesh["cp"].get_group()
    else:
        cp_group = None

    return partial(
        async_safe_ce,
        loss_scaling_factor=loss_scaling_factor,
        dp_group=dp_group,
        cp_group=cp_group,
        ignore_index=IGNORE_INDEX,
    )


def init_cosmos_rl_model(config, is_train=True, device="cuda"):
    if config.policy.parallelism.dp_replicate_size == -1:
        # recompute it
        world_size = int(os.environ.get("WORLD_SIZE", 1))
        config.policy.parallelism.dp_replicate_size = (
            world_size // config.policy.parallelism.dp_shard_size
        )
        print(
            f"Recomputed dp_replicate_size: {config.policy.parallelism.dp_replicate_size} given world_size: {world_size} and dp_shard_size: {config.policy.parallelism.dp_shard_size}"
        )

    model = ModelRegistry.build_model(config)
    # if config.policy.model_gradient_checkpointing:
    #     apply_ac(model, config.policy.model_name_or_path)
    # init parallel_dims
    parallel_dims: ParallelDims = ParallelDims.from_config(
        parallesim_config=config.policy.parallelism
    )
    parallel_dims.build_mesh(device_type=device)

    print(f"parallel_dims: {parallel_dims}")

    parallelize_fn, _ = model.parallelize_fn
    if is_train:
        loss_fn = pp_loss_fn(config, parallel_dims)
    else:
        loss_fn = None
    pp_scheduler, pp_scheduler_val = parallelize_fn(
        model, parallel_dims, config, pp_loss_fn=loss_fn
    )
    assert pp_scheduler is None, "pp_scheduler should be None"
    assert pp_scheduler_val is None, "pp_scheduler_val should be None"
    if not config.train.fsdp_offload:
        model.to_empty(device=device)
    model.post_to_empty_hook(config)

    torch.cuda.empty_cache()
    model.load_hf_weights(
        config.policy.model_name_or_path,
        parallel_dims,
        device,
    )
    return [model], pp_scheduler, parallel_dims, loss_fn


class Config:
    def __init__(self, config_dict):
        for k, v in config_dict.items():
            if isinstance(v, dict):
                v = Config(v)
            setattr(self, k, v)


def create_assistant_tokens_mask(tokens, processor):  # Qwen2 based model
    if isinstance(tokens, torch.Tensor) and tokens.ndim == 2:
        mask = torch.stack(
            [
                create_assistant_tokens_mask(tokens[i], processor)
                for i in range(tokens.shape[0])
            ]
        )
        assert mask.shape == tokens.shape
        return mask
    np_tokens = (
        tokens.cpu().numpy() if isinstance(tokens, torch.Tensor) else np.array(tokens)
    )
    assert np_tokens.ndim == 1
    # Constants defining bos, eos and fixed offsets.
    BOS_TOKEN = "<|im_start|>"
    EOS_TOKEN = "<|im_end|>"
    ROLE = "assistant"
    # Offsets: skip the bos + "assistant\n" (always 3 tokens) and include the eos (+1) for supervision
    START_OFFSET = 3
    END_OFFSET = 1

    # Retrieve token IDs for the markers and the role.
    bos_token_id = processor.tokenizer.convert_tokens_to_ids(BOS_TOKEN)
    eos_token_id = processor.tokenizer.convert_tokens_to_ids(EOS_TOKEN)
    role_id = processor.tokenizer.convert_tokens_to_ids(ROLE)

    # Locate all positions where the start and end markers appear.
    start_indices = np.where(np_tokens == bos_token_id)[0]
    end_indices = np.where(np_tokens == eos_token_id)[0]

    # Initialize the mask with False values.
    masks = np.zeros_like(np_tokens, dtype=bool)
    assert len(start_indices) == len(
        end_indices
    ), "remember to set add_generation_prompt=False in apply_chat_template"
    # For each pair of bos/eos, check if the role is 'assistant'
    # and apply the mask accordingly.
    for start, end in zip(start_indices, end_indices):
        if np_tokens[start + 1] == role_id:
            print(
                f"start: {start}, end: {end}, role_id: {role_id}, np_tokens[start + 1]: {np_tokens[start + 1]}"
            )
            # Mask tokens from after the assistant header (start+3) to include the end marker (end+1)
            masks[start + START_OFFSET : end + END_OFFSET] = True
    assert masks.shape == np_tokens.shape
    print(f"masks: {masks}")
    if isinstance(tokens, torch.Tensor):
        return torch.from_numpy(masks)
    else:
        return masks.tolist()


def create_debug_input(model_name_or_path):
    processor = AutoProcessor.from_pretrained(model_name_or_path)
    messages = [
        {
            "role": "user",
            "content": [
                {
                    "type": "image",
                    "image": "https://qianwen-res.oss-cn-beijing.aliyuncs.com/Qwen-VL/assets/demo.jpeg",
                    "min_pixels": 1024 * 28 * 28,
                    "max_pixels": 1024 * 28 * 28,
                },
                {"type": "text", "text": "Describe the image caption."},
            ],
        },
        {
            "role": "assistant",
            "content": [
                {
                    "type": "text",
                    "text": "The image captures a serene and joyful moment on a sandy beach at sunset. The sky is a soft gradient of warm colors, transitioning from a pale blue at the horizon to a deeper, golden hue near the horizon line. The ocean waves gently roll onto the shore, creating a tranquil and picturesque backdrop.\nIn the foreground, a person is sitting on the sand, facing a light-colored dog, likely a Labrador Retriever. The person is dressed in a plaid shirt and dark pants, with their legs crossed and one hand resting on their knee. The dog is wearing a harness with a colorful collar and is sitting on its haunches, facing the person. The dog's tail is wagging, and it appears to be engaged in a playful interaction with the person, possibly giving a high-five or pawing at their hand.",
                },
            ],
        },
    ]

    # Preparation for inference
    inputs = processor.apply_chat_template(
        messages,
        tokenize=True,
        add_generation_prompt=False,
        return_dict=True,
        return_tensors="pt",
    )
    for k, v in inputs.items():
        print(f"k: {k}, type: {type(v)}")
        if isinstance(v, torch.Tensor):
            inputs[k] = v.to("cuda")

    labels = inputs["input_ids"].clone()
    token_mask = create_assistant_tokens_mask(inputs["input_ids"], processor)
    labels[~token_mask] = IGNORE_INDEX
    return {
        "input_ids": inputs["input_ids"],
        "attention_mask": inputs["attention_mask"],
        "pixel_values": inputs["pixel_values"],
        "image_grid_thw": inputs["image_grid_thw"],
        "labels": labels,
    }


class TestCosmosHfPrecision(unittest.TestCase):
    def test_cosmos_hf_precision(self):
        model_name = "Qwen/Qwen3-VL-30B-A3B-Instruct"
        # ================================
        # create config
        # ================================
        config_dict = {
            "policy": {
                "model_name_or_path": model_name,
                "model_max_length": 40960,
                "parallelism": {
                    "tp_size": 4,
                    "cp_size": 1,
                    "ep_size": 1,
                    "dp_shard_size": -1,
                    "dp_replicate_size": 1,
                    "pp_size": 1,
                    "world_size": int(os.environ.get("WORLD_SIZE", 1)),
                    "pp_dynamic_shape": False,
                },
                "lora": None,
                "enable_liger_kernel": False,
                "trainable_map": None,
                "model_gradient_checkpointing": True,
            },
            "train": {
                "fsdp_offload": False,
                "compile": False,
                "master_dtype": None,
                "param_dtype": "bfloat16",
                "fsdp_reduce_dtype": "float32",
                "fsdp_reshard_after_forward": "default",
                "param_torch_dtype": torch.bfloat16,
                "train_policy": {"mini_batch": 1},
                "fp8": {"enable_fp8": False},
                "async_tp_enabled": False,
                # "sequence_packing": True,
            },
            "rollout": {
                "multi_turn_config": {
                    "enable": False,
                    "custom_chat_template_path": None,
                },
            },
        }
        config = Config(config_dict)

        # ================================
        # create data packer
        # ================================
        data_packer = Qwen3_VL_DataPacker()
        data_packer.setup(
            config=config, tokenizer=AutoTokenizer.from_pretrained(model_name)
        )

        if int(os.environ.get("WORLD_SIZE", 1)) > 1:
            if not torch.distributed.is_initialized():
                torch.distributed.init_process_group(backend="cuda:nccl,cpu:gloo")
                torch.cuda.set_device(torch.distributed.get_rank())

        # ================================
        # create data batch
        # ================================
        data_batch = create_debug_input(model_name_or_path=model_name)
        # fill in position_ids
        position_ids, _ = data_packer._get_rope_index(
            input_ids=data_batch["input_ids"],
            image_grid_thw=data_batch.get("image_grid_thw", None),
            video_grid_thw=data_batch.get("video_grid_thw", None),
            attention_mask=data_batch.get("attention_mask", None),
        )
        if position_ids is not None:
            data_batch["position_ids"] = position_ids.permute(1, 0, 2).contiguous()
            print(f"position_ids: {data_batch['position_ids'].shape}")

        for k, v in data_batch.items():
            data_batch[k] = v.to("cuda")

        # compute logits for hf model
        hf_model = Qwen3VLMoeForConditionalGeneration.from_pretrained(
            model_name,
            device_map="auto",
            dtype=torch.bfloat16,
            attn_implementation="flash_attention_2",
        )

        hf_model.eval()
        with torch.no_grad():
            logits_hf = hf_model(
                data_batch["input_ids"],
                pixel_values=data_batch["pixel_values"],
                attention_mask=data_batch["attention_mask"],
                image_grid_thw=data_batch["image_grid_thw"],
                use_cache=False,
                return_dict=True,
            ).logits
            print(f"before logits_hf: {logits_hf.shape}")
        del hf_model
        torch.cuda.empty_cache()
        ce_loss_hf = torch.nn.functional.cross_entropy(
            input=logits_hf[:, :-1].flatten(0, 1),
            target=data_batch["labels"][:, 1:].flatten(0, 1),
            ignore_index=-100,
            reduction="none",
        )
        ce_loss_hf_mean = ce_loss_hf.mean()
        logits_hf = logits_hf[:, -1, :]

        cosmos_model_list, _, _, _ = init_cosmos_rl_model(config, device="cuda")
        cosmos_model = cosmos_model_list[0]
        cosmos_model.eval()
        with torch.no_grad():
            logits_cosmos_rl = cosmos_model(
                data_batch["input_ids"],
                pixel_values=data_batch["pixel_values"],
                attention_mask=data_batch["attention_mask"],
                image_grid_thw=data_batch["image_grid_thw"],
                position_ids=data_batch.get("position_ids", None),
            )
        del cosmos_model

        ce_loss = torch.nn.functional.cross_entropy(
            input=logits_cosmos_rl[:, :-1].flatten(0, 1),
            target=data_batch["labels"][:, 1:].flatten(0, 1),
            ignore_index=-100,
            reduction="none",
        )
        cs_loss_mean = ce_loss.mean()
        logits_cosmos_rl = logits_cosmos_rl[:, -1, :]
        torch.cuda.empty_cache()

        if torch.distributed.get_rank() == 0:
            max_index_hf = logits_hf.argmax(dim=-1)
            max_index_cosmos_rl = logits_cosmos_rl.argmax(dim=-1)
            diff_loss_mean = (ce_loss_hf_mean - cs_loss_mean).abs()
            diff_logits = (logits_cosmos_rl - logits_hf).abs()
            print(
                f"ce_loss_hf_mean: {ce_loss_hf_mean} | cs_loss_mean: {cs_loss_mean} | diff_loss_mean: {diff_loss_mean}"
            )
            print(
                f"logits_hf: {logits_hf} | logits_cosmos_rl: {logits_cosmos_rl} | diff_logits: {diff_logits}"
            )
            print(
                f"max_index_hf: {max_index_hf} | max_index_cosmos_rl: {max_index_cosmos_rl}"
            )
            assert diff_loss_mean < 1e-3
            assert max_index_hf == max_index_cosmos_rl


# torchrun --nproc_per_node=4 test_qwen3_vl_moe.py
if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Callable, Tuple, List
from dataclasses import dataclass
import torch
import torch.nn as nn
import torch.nn.functional as F
from transformers.activations import ACT2FN
from transformers import AutoConfig
import cosmos_rl.policy.kernel.modeling_utils as modeling_utils
from cosmos_rl.policy.kernel.norm import RMSNorm
from cosmos_rl.policy.kernel.fused import MLPActMulFunc


@dataclass
class Qwen2_5_VL_Encoder_Args:
    depth: int
    hidden_size: int
    hidden_act: str
    intermediate_size: int
    n_heads: int
    in_channels: int
    patch_size: int
    spatial_merge_size: int
    temporal_patch_size: int
    tokens_per_second: int
    window_size: int
    fullatt_block_indexes: List[int]
    out_hidden_size: int
    norm_type: str = "rmsnorm"
    norm_eps: float = 1e-6
    hf_config: AutoConfig = None


class Qwen2_5_VLMLP(nn.Module):
    def __init__(self, config: Qwen2_5_VL_Encoder_Args, bias: bool = False):
        super().__init__()
        self.hidden_size = config.hidden_size
        self.intermediate_size = config.intermediate_size
        self.gate_proj = nn.Linear(self.hidden_size, self.intermediate_size, bias=bias)
        self.up_proj = nn.Linear(self.hidden_size, self.intermediate_size, bias=bias)
        self.down_proj = nn.Linear(self.intermediate_size, self.hidden_size, bias=bias)
        self.act_mul_func = MLPActMulFunc(ACT2FN[config.hidden_act])

    def forward(self, hidden_state):
        return self.down_proj(
            self.act_mul_func(self.gate_proj(hidden_state), self.up_proj(hidden_state))
        )


class Qwen2_5_VisionPatchEmbed(nn.Module):
    def __init__(
        self,
        patch_size: int = 14,
        temporal_patch_size: int = 2,
        in_channels: int = 3,
        embed_dim: int = 1152,
    ) -> None:
        super().__init__()
        self.patch_size = patch_size
        self.temporal_patch_size = temporal_patch_size
        self.in_channels = in_channels
        self.embed_dim = embed_dim

        kernel_size = [temporal_patch_size, patch_size, patch_size]
        self.proj = nn.Conv3d(
            in_channels,
            embed_dim,
            kernel_size=kernel_size,
            stride=kernel_size,
            bias=False,
        )

    def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
        target_dtype = self.proj.weight.dtype
        hidden_states = hidden_states.view(
            -1,
            self.in_channels,
            self.temporal_patch_size,
            self.patch_size,
            self.patch_size,
        )
        hidden_states = self.proj(hidden_states.to(dtype=target_dtype)).view(
            -1, self.embed_dim
        )
        return hidden_states


class Qwen2_5_VisionRotaryEmbedding(nn.Module):
    def __init__(self, dim: int, theta: float = 10000.0) -> None:
        super().__init__()
        self.theta = theta
        self.dim = dim
        self.reset_inv_freq()

    def reset_inv_freq(self, device: torch.device = None):
        inv_freq = 1.0 / (
            self.theta ** (torch.arange(0, self.dim, 2, dtype=torch.float) / self.dim)
        )
        inv_freq = inv_freq.to(device) if device is not None else inv_freq
        if not hasattr(self, "inv_freq"):
            self.register_buffer("inv_freq", inv_freq, persistent=False)
        else:
            self.inv_freq.to(torch.float32)
            with torch.no_grad():
                self.inv_freq.data.copy_(inv_freq)

    def forward(self, seqlen: int, device: torch.device = None) -> torch.Tensor:
        if self.inv_freq.dtype != torch.float32:
            self.reset_inv_freq(device=device)
            assert (
                self.inv_freq.dtype == torch.float32
            ), "inv_freq dtype should be float32"
        seq = torch.arange(
            seqlen, device=self.inv_freq.device, dtype=self.inv_freq.dtype
        )
        freqs = torch.outer(seq, self.inv_freq)
        return freqs


class Qwen2_5_VLPatchMerger(nn.Module):
    def __init__(self, config: Qwen2_5_VL_Encoder_Args) -> None:
        super().__init__()
        self.hidden_size = config.hidden_size * (config.spatial_merge_size**2)
        self.ln_q = RMSNorm(
            config.hidden_size,
            config.norm_eps,
            casting_mode=config.hf_config.model_type,
        )
        self.mlp = nn.Sequential(
            nn.Linear(self.hidden_size, self.hidden_size),
            nn.GELU(),  # This is fixed to GELU according to the original implementation
            nn.Linear(self.hidden_size, config.out_hidden_size),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.mlp(self.ln_q(x).view(-1, self.hidden_size))
        return x


def rotate_half(x):
    """Rotates half the hidden dims of the input."""
    x1 = x[..., : x.shape[-1] // 2]
    x2 = x[..., x.shape[-1] // 2 :]
    return torch.cat((-x2, x1), dim=-1)


if os.environ.get("COSMOS_USE_HF_IMPL", "0").lower() in ["1", "true"]:

    def apply_rotary_pos_emb_vision(
        q: torch.Tensor, k: torch.Tensor, cos: torch.Tensor, sin: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor]:
        orig_q_dtype = q.dtype
        orig_k_dtype = k.dtype
        q, k = q.float(), k.float()
        cos, sin = cos.unsqueeze(-2).float(), sin.unsqueeze(-2).float()
        q_embed = (q * cos) + (rotate_half(q) * sin)
        k_embed = (k * cos) + (rotate_half(k) * sin)
        q_embed = q_embed.to(orig_q_dtype)
        k_embed = k_embed.to(orig_k_dtype)
        return q_embed, k_embed
else:

    def apply_rotary_pos_emb_vision(
        q: torch.Tensor, k: torch.Tensor, cos: torch.Tensor, sin: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Apply rotary position embedding to the query and key tensors.
        """
        cos = cos.chunk(2, dim=-1)[0].contiguous()
        sin = sin.chunk(2, dim=-1)[0].contiguous()
        q_embed = modeling_utils.apply_rotary_emb(
            q.float(), cos.float(), sin.float()
        ).type_as(q)
        k_embed = modeling_utils.apply_rotary_emb(
            k.float(), cos.float(), sin.float()
        ).type_as(k)
        return q_embed, k_embed


class Qwen2_5_VLVisionAttention(nn.Module):
    def __init__(self, dim: int, num_heads: int = 16) -> None:
        super().__init__()
        self.num_heads = num_heads
        self.qkv = nn.Linear(dim, dim * 3, bias=True)
        self.proj = nn.Linear(dim, dim)
        self.attention_dropout = 0.0
        self.attn_func = modeling_utils.flash_attn_varlen_func

    def forward(
        self,
        hidden_states: torch.Tensor,
        cu_seqlens: torch.Tensor,
        position_embeddings: torch.Tensor,
        max_seqlen: int,
    ) -> torch.Tensor:
        seq_length = hidden_states.shape[0]
        q, k, v = (
            self.qkv(hidden_states)
            .reshape(seq_length, 3, self.num_heads, -1)
            .permute(1, 0, 2, 3)
            .unbind(0)
        )
        cos, sin = position_embeddings

        q, k = apply_rotary_pos_emb_vision(q.unsqueeze(0), k.unsqueeze(0), cos, sin)
        q = q.squeeze(0)
        k = k.squeeze(0)

        input_dtype = q.dtype
        if input_dtype == torch.float32:
            if torch.is_autocast_enabled():
                target_dtype = torch.get_autocast_gpu_dtype()
            else:
                raise ValueError("Flash attention only supports float32 input")
            q = q.to(target_dtype)
            k = k.to(target_dtype)
            v = v.to(target_dtype)

        attn_output = self.attn_func(
            q,
            k,
            v,
            cu_seqlens_q=cu_seqlens,
            cu_seqlens_k=cu_seqlens,
            max_seqlen_q=max_seqlen,
            max_seqlen_k=max_seqlen,
            causal=False,
        ).reshape(seq_length, -1)
        attn_output = self.proj(attn_output)
        return attn_output


class Qwen2_5_VLVisionBlock(nn.Module):
    def __init__(self, config: Qwen2_5_VL_Encoder_Args) -> None:
        super().__init__()
        self.norm1 = RMSNorm(
            config.hidden_size,
            config.norm_eps,
            casting_mode=config.hf_config.model_type,
        )
        self.norm2 = RMSNorm(
            config.hidden_size,
            config.norm_eps,
            casting_mode=config.hf_config.model_type,
        )
        self.attn = Qwen2_5_VLVisionAttention(
            config.hidden_size, num_heads=config.n_heads
        )
        self.mlp = Qwen2_5_VLMLP(
            config,
            bias=True,  # This is fixed to True according to the original implementation
        )

    def forward(
        self, hidden_states, cu_seqlens, position_embeddings, max_seqlen
    ) -> torch.Tensor:
        hidden_states = hidden_states + self.attn(
            self.norm1(hidden_states),
            cu_seqlens=cu_seqlens,
            position_embeddings=position_embeddings,
            max_seqlen=max_seqlen,
        )
        hidden_states = hidden_states + self.mlp(self.norm2(hidden_states))
        return hidden_states


class Qwen2_5_VisionTransformerPretrainedModel(nn.Module):
    def __init__(self, config) -> None:
        super().__init__()
        self.config = config
        self.spatial_merge_size = config.spatial_merge_size
        self.patch_size = config.patch_size
        self.fullatt_block_indexes = config.fullatt_block_indexes
        self.window_size = config.window_size
        self.spatial_merge_unit = self.spatial_merge_size * self.spatial_merge_size

        self.patch_embed = Qwen2_5_VisionPatchEmbed(
            patch_size=config.patch_size,
            temporal_patch_size=config.temporal_patch_size,
            in_channels=config.in_channels,
            embed_dim=config.hidden_size,
        )

        head_dim = config.hidden_size // config.n_heads
        self.rotary_pos_emb = Qwen2_5_VisionRotaryEmbedding(head_dim // 2)

        self.blocks = torch.nn.ModuleDict()
        for layer_id in range(config.depth):
            self.blocks[str(layer_id)] = Qwen2_5_VLVisionBlock(config)
        self.merger = Qwen2_5_VLPatchMerger(config=config)

    def rot_pos_emb(self, grid_thw):
        pos_ids = []
        for t, h, w in grid_thw:
            hpos_ids = torch.arange(h).unsqueeze(1).expand(-1, w)
            hpos_ids = hpos_ids.reshape(
                h // self.spatial_merge_size,
                self.spatial_merge_size,
                w // self.spatial_merge_size,
                self.spatial_merge_size,
            )
            hpos_ids = hpos_ids.permute(0, 2, 1, 3)
            hpos_ids = hpos_ids.flatten()

            wpos_ids = torch.arange(w).unsqueeze(0).expand(h, -1)
            wpos_ids = wpos_ids.reshape(
                h // self.spatial_merge_size,
                self.spatial_merge_size,
                w // self.spatial_merge_size,
                self.spatial_merge_size,
            )
            wpos_ids = wpos_ids.permute(0, 2, 1, 3)
            wpos_ids = wpos_ids.flatten()
            pos_ids.append(torch.stack([hpos_ids, wpos_ids], dim=-1).repeat(t, 1))
        pos_ids = torch.cat(pos_ids, dim=0)
        max_grid_size = grid_thw[:, 1:].max()
        rotary_pos_emb_full = self.rotary_pos_emb(max_grid_size)
        rotary_pos_emb = rotary_pos_emb_full[pos_ids].flatten(1)
        return rotary_pos_emb

    def get_window_index(self, grid_thw):
        window_index: list = []
        cu_window_seqlens: list = [0]
        window_index_id = 0
        vit_merger_window_size = (
            self.window_size // self.spatial_merge_size // self.patch_size
        )

        for grid_t, grid_h, grid_w in grid_thw:
            llm_grid_h, llm_grid_w = (
                grid_h // self.spatial_merge_size,
                grid_w // self.spatial_merge_size,
            )
            index = torch.arange(grid_t * llm_grid_h * llm_grid_w).reshape(
                grid_t, llm_grid_h, llm_grid_w
            )
            pad_h = vit_merger_window_size - llm_grid_h % vit_merger_window_size
            pad_w = vit_merger_window_size - llm_grid_w % vit_merger_window_size
            num_windows_h = (llm_grid_h + pad_h) // vit_merger_window_size
            num_windows_w = (llm_grid_w + pad_w) // vit_merger_window_size
            index_padded = F.pad(index, (0, pad_w, 0, pad_h), "constant", -100)
            index_padded = index_padded.reshape(
                grid_t,
                num_windows_h,
                vit_merger_window_size,
                num_windows_w,
                vit_merger_window_size,
            )
            index_padded = index_padded.permute(0, 1, 3, 2, 4).reshape(
                grid_t,
                num_windows_h * num_windows_w,
                vit_merger_window_size,
                vit_merger_window_size,
            )
            seqlens = (index_padded != -100).sum([2, 3]).reshape(-1)
            index_padded = index_padded.reshape(-1)
            index_new = index_padded[index_padded != -100]
            window_index.append(index_new + window_index_id)
            cu_seqlens_tmp = (
                seqlens.cumsum(0) * self.spatial_merge_unit + cu_window_seqlens[-1]
            )
            cu_window_seqlens.extend(cu_seqlens_tmp.tolist())
            window_index_id += (grid_t * llm_grid_h * llm_grid_w).item()
        window_index = torch.cat(window_index, dim=0)

        return window_index, cu_window_seqlens

    def forward(
        self, hidden_states: torch.Tensor, grid_thw: torch.Tensor
    ) -> torch.Tensor:
        """
        Args:
            hidden_states (`torch.Tensor` of shape `(seq_len, hidden_size)`):
                The final hidden states of the model.
            grid_thw (`torch.Tensor` of shape `(num_images_or_videos, 3)`):
                The temporal, height and width of feature shape of each image in LLM.

        Returns:
            `torch.Tensor`: hidden_states.
        """
        hidden_states = self.patch_embed(hidden_states)
        rotary_pos_emb = self.rot_pos_emb(grid_thw)
        window_index, cu_window_seqlens = self.get_window_index(grid_thw)
        cu_window_seqlens = torch.tensor(
            cu_window_seqlens,
            device=hidden_states.device,
            dtype=grid_thw.dtype if torch.jit.is_tracing() else torch.int32,
        )
        cu_window_seqlens = torch.unique_consecutive(cu_window_seqlens)

        seq_len, _ = hidden_states.size()
        hidden_states = hidden_states.reshape(
            seq_len // self.spatial_merge_unit, self.spatial_merge_unit, -1
        )
        hidden_states = hidden_states[window_index, :, :]
        hidden_states = hidden_states.reshape(seq_len, -1)
        rotary_pos_emb = rotary_pos_emb.reshape(
            seq_len // self.spatial_merge_unit, self.spatial_merge_unit, -1
        )
        rotary_pos_emb = rotary_pos_emb[window_index, :, :]
        rotary_pos_emb = rotary_pos_emb.reshape(seq_len, -1)
        emb = torch.cat((rotary_pos_emb, rotary_pos_emb), dim=-1)
        position_embeddings = (emb.cos(), emb.sin())

        cu_seqlens = torch.repeat_interleave(
            grid_thw[:, 1] * grid_thw[:, 2], grid_thw[:, 0]
        ).cumsum(
            dim=0,
            # Select dtype based on the following factors:
            #  - FA2 requires that cu_seqlens_q must have dtype int32
            #  - torch.onnx.export requires that cu_seqlens_q must have same dtype as grid_thw
            # See https://github.com/huggingface/transformers/pull/34852 for more information
            dtype=grid_thw.dtype if torch.jit.is_tracing() else torch.int32,
        )
        cu_seqlens = F.pad(cu_seqlens, (1, 0), value=0)

        with torch.no_grad():
            max_window_seqlen = (
                (cu_window_seqlens[1:] - cu_window_seqlens[:-1]).max().item()
            )
            max_seqlen = (cu_seqlens[1:] - cu_seqlens[:-1]).max().item()

        for layer_num, blk in self.blocks.items():
            if int(layer_num) in self.fullatt_block_indexes:
                cu_seqlens_now = cu_seqlens
                max_seqlen_now = max_seqlen
            else:
                cu_seqlens_now = cu_window_seqlens
                max_seqlen_now = max_window_seqlen

            if (
                hasattr(blk, "_gradient_checkpointing_enabled")
                and blk._gradient_checkpointing_enabled
            ):
                hidden_states = torch.utils.checkpoint.checkpoint(
                    blk,
                    hidden_states,
                    cu_seqlens_now,
                    position_embeddings,
                    max_seqlen_now,
                    use_reentrant=False,
                )
            else:
                hidden_states = blk(
                    hidden_states,
                    cu_seqlens=cu_seqlens_now,
                    position_embeddings=position_embeddings,
                    max_seqlen=max_seqlen_now,
                )

        hidden_states = self.merger(hidden_states)
        reverse_indices = torch.argsort(window_index)
        hidden_states = hidden_states[reverse_indices, :]

        return hidden_states

    def _get_nparams_and_flops_fn(self) -> Callable[[int], tuple[int, int]]:
        nparams = sum(p.numel() for p in self.parameters())
        nparams_embedding = sum(
            sum(p.numel() for p in m.parameters())
            for m in self.children()
            if isinstance(m, nn.Embedding)
        )

        # Reasoning behind the factor of 12 for the self-attention part of the formula:
        # 1. each self-attention has 2 matmul in the forward and 4 in the backward (6)
        # 2. the flash attention does 1 more matmul recomputation in the backward
        #    but recomputation should not be counted in calculating MFU           (+0)
        # 3. each matmul performs 1 multiplication and 1 addition                 (*2)
        # 4. we follow the convention and do not account for sparsity in causal attention
        layers, heads, head_dim = (
            len(self.blocks),
            self.config.n_heads,
            self.config.hidden_size // self.config.n_heads,
        )
        return lambda seq_len: (
            nparams,
            6 * (nparams - nparams_embedding)
            + 12 * layers * heads * head_dim * seq_len,
        )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import DataPacker
from .decoder_only_llm_data_packer import DecoderOnlyLLMDataPacker
from .qwen2_5_vlm_data_packer import Qwen2_5_VLM_DataPacker
from .qwen3_vl_data_packer import Qwen3_VL_DataPacker
from .hf_vlm_data_packer import HFVLMDataPacker
from .base import worker_entry_parser

__all__ = [
    "DataPacker",
    "DecoderOnlyLLMDataPacker",
    "Qwen2_5_VLM_DataPacker",
    "Qwen3_VL_DataPacker",
    "HFVLMDataPacker",
    "worker_entry_parser",
]



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import heapq
from cosmos_rl.utils.logging import logger
import cosmos_rl.utils.distributed as dist_util
from cosmos_rl.utils.distributed import HighAvailabilitylNccl
import torch


def karmarkar_karp(seqlen_list: list[int], k_partitions: int, equal_size: bool):
    # Karmarkar-Karp differencing method for k-partition problem: https://en.wikipedia.org/wiki/Largest_differencing_method
    # Implementation Reference: https://github.com/volcengine/verl/blob/main/verl/utils/seqlen_balancing.py#L27
    class Set:
        def __init__(self) -> None:
            self.sum = 0
            self.items = []

        def add(self, idx: int, val: int):
            self.items.append((idx, val))
            self.sum += val

        def merge(self, other):
            for idx, val in other.items:
                self.items.append((idx, val))
                self.sum += val

        def __lt__(self, other):
            if self.sum != other.sum:
                return self.sum < other.sum
            if len(self.items) != len(other.items):
                return len(self.items) < len(other.items)
            return self.items < other.items

    class State:
        def __init__(self, items: list[tuple[int, int]], k: int) -> None:
            self.k = k
            # sets should always be decreasing order
            self.sets = [Set() for _ in range(k)]
            assert len(items) in [1, k], f"{len(items)} not in [1, {k}]"
            for i, (idx, seqlen) in enumerate(items):
                self.sets[i].add(idx=idx, val=seqlen)
            self.sets = sorted(self.sets, reverse=True)

        def get_partitions(self):
            partitions = []
            for i in range(len(self.sets)):
                cur_partition = []
                for idx, _ in self.sets[i].items:
                    cur_partition.append(idx)
                partitions.append(cur_partition)
            return partitions

        def merge(self, other):
            for i in range(self.k):
                self.sets[i].merge(other.sets[self.k - 1 - i])
            self.sets = sorted(self.sets, reverse=True)

        @property
        def spread(self) -> int:
            return self.sets[0].sum - self.sets[-1].sum

        def __lt__(self, other):
            # least heap, let the state with largest spread to be popped first,
            # if the spread is the same, let the state who has the largest set
            # to be popped first.
            if self.spread != other.spread:
                return self.spread > other.spread
            return self.sets[0] > other.sets[0]

        def __repr__(self) -> str:
            repr_str = "["
            for i in range(self.k):
                if i > 0:
                    repr_str += ","
                repr_str += "{"
                for j, (_, seqlen) in enumerate(self.sets[i].items):
                    if j > 0:
                        repr_str += ","
                    repr_str += str(seqlen)
                repr_str += "}"
            repr_str += "]"
            return repr_str

    sorted_seqlen_list = sorted([(seqlen, i) for i, seqlen in enumerate(seqlen_list)])
    states_pq = []
    if equal_size:
        assert (
            len(seqlen_list) % k_partitions == 0
        ), f"{len(seqlen_list)} % {k_partitions} != 0"
        for offset in range(0, len(sorted_seqlen_list), k_partitions):
            items = []
            for i in range(k_partitions):
                seqlen, idx = sorted_seqlen_list[offset + i]
                items.append((idx, seqlen))
            heapq.heappush(states_pq, State(items=items, k=k_partitions))
    else:
        for seqlen, idx in sorted_seqlen_list:
            heapq.heappush(states_pq, State(items=[(idx, seqlen)], k=k_partitions))

    while len(states_pq) > 1:
        state0 = heapq.heappop(states_pq)
        state1 = heapq.heappop(states_pq)
        # merge states
        state0.merge(state1)
        heapq.heappush(states_pq, state0)

    final_state = states_pq[0]
    partitions = final_state.get_partitions()
    if equal_size:
        for i, partition in enumerate(partitions):
            assert len(partition) * k_partitions == len(
                seqlen_list
            ), f"{len(partition)} * {k_partitions} != {len(seqlen_list)}"
    return partitions


# Refer to: https://github.com/volcengine/verl/blob/main/verl/utils/seqlen_balancing.py#L151
def get_seqlen_balanced_partitions(
    seqlen_list: list[int], k_partitions: int, equal_size: bool
):
    """
    Calculates partitions of indices from seqlen_list such that the sum of sequence lengths
    in each partition is balanced. Uses the Karmarkar-Karp differencing method.

    This is useful for balancing workload across devices or batches, especially when
    dealing with variable sequence lengths.

    Args:
        seqlen_list (List[int]): A list of sequence lengths for each item.
        k_partitions (int): The desired number of partitions.
        equal_size (bool): If True, ensures that each partition has the same number of items.
                           Requires len(seqlen_list) to be divisible by k_partitions.
                           If False, partitions can have varying numbers of items, focusing
                           only on balancing the sum of sequence lengths.

    Returns:
        List[List[int]]: A list containing k_partitions lists. Each inner list contains the
                         original indices of the items assigned to that partition. The indices
                         within each partition list are sorted.

    Raises:
        AssertionError: If len(seqlen_list) < k_partitions.
        AssertionError: If equal_size is True and len(seqlen_list) is not divisible by k_partitions.
        AssertionError: If any resulting partition is empty.
    """
    assert (
        len(seqlen_list) >= k_partitions
    ), f"number of items:[{len(seqlen_list)}] < k_partitions:[{k_partitions}]"

    def _check_and_sort_partitions(partitions):
        assert len(partitions) == k_partitions, f"{len(partitions)} != {k_partitions}"
        seen_idx = set()
        sorted_partitions = [None] * k_partitions
        for i, partition in enumerate(partitions):
            assert len(partition) > 0, f"the {i}-th partition is empty"
            for idx in partition:
                seen_idx.add(idx)
            sorted_partitions[i] = sorted(partition)
        assert seen_idx == set(range(len(seqlen_list)))
        return sorted_partitions

    partitions = karmarkar_karp(
        seqlen_list=seqlen_list, k_partitions=k_partitions, equal_size=equal_size
    )
    return _check_and_sort_partitions(partitions)


def ceildiv(a, b):
    return -(a // -b)


# Modified based on: https://github.com/volcengine/verl/blob/main/verl/utils/seqlen_balancing.py#L251
def rearrange_mini_batches(
    batch,
    seq_len_effective: list[int],
    max_token_len: int,
    dp_group=None,
    same_mini_num_in_dp=True,
    min_num_mini_batch=None,
    use_dynamic_bsz_balance=True,
    ddp_comm: HighAvailabilitylNccl = None,
):
    """
    Split a batch into mini-batches by total token count, with optional DP sync and padding.

    Args:
        batch (TensorDict): mini-batch data to be split into mini-batches.
        seq_len_effective (List[int]): list of effective sequence lengths for each item in the batch.
        max_token_len (int): max sum of effective tokens per mini-batch.
        dp_group (optional): torch.distributed group for data-parallel sync.
        same_mini_num_in_dp (bool): if True and dp_group set, pad all ranks to the same count.
        min_num_mini_batch (int, optional): force at least this many splits (pads empty ones).
        use_dynamic_bsz_balance (bool, optional): balance the computational workload between mini-batches.
        ddp_comm: HighAvailabilitylNccl = None, optional DDP communicator for distributed operations.
    Returns:
        List[List[Any]]: the mini-batches after rearrangement.
        List[List[int]]: index lists mapping each mini-batch back to original positions.
    """
    # this is per local mini_bsz
    max_seq_len = max(seq_len_effective)
    if max_token_len < max_seq_len:
        logger.warning(
            f"max_token_len[{max_token_len}] < max_seq_len[{max_seq_len}], "
            f"setting max_token_len = max_seq_len"
        )
        max_token_len = max_seq_len
    total_seqlen = sum(seq_len_effective)
    num_mini_batches = min(len(seq_len_effective), ceildiv(total_seqlen, max_token_len))
    if min_num_mini_batch is not None:
        num_mini_batches = max(min_num_mini_batch, num_mini_batches)
    if same_mini_num_in_dp:
        num_mini_batches = max(
            dist_util.all_gather_object_cpu(num_mini_batches, group=dp_group)
        )

    if ddp_comm is not None:
        num_mini_batches_tensor = torch.tensor(num_mini_batches, device="cuda")
        ddp_comm.allreduce(
            num_mini_batches_tensor,
            num_mini_batches_tensor,
            op=torch.distributed.ReduceOp.MAX,
        )
        num_mini_batches = num_mini_batches_tensor.item()

    assert num_mini_batches <= len(seq_len_effective)
    mini_bsz_idx = get_seqlen_balanced_partitions(
        seq_len_effective, num_mini_batches, equal_size=False
    )

    if use_dynamic_bsz_balance:
        # Use the sum of squared sequence lengths to approximate attention computation workload
        mini_bsz_idx.sort(
            key=lambda partition: (
                sum(seq_len_effective[idx] ** 2 for idx in partition),
                min(partition) if partition else 0,
            ),
            reverse=True,
        )

    mini_batches = []

    for partition in mini_bsz_idx:
        curr_mini_batch = []
        for idx in partition:
            curr_mini_batch.append(batch[idx])
        mini_batches.append(curr_mini_batch)

    return mini_batches, mini_bsz_idx
